/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora mathematical functions evaluation
 *
 *  Copyright (C) 2009--2013 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsl/gsl_spline.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_sort_double.h>
#include <gsl/gsl_sort_vector_double.h>

#include "wasora.h"
#include "thirdparty/kdtree.h"

// setea las variables que son argumento de una funcion al valor pedido en el vector x
void wasora_set_function_args(function_t *function, double *x) {
  int i;
  
  if (function->var_argument != NULL) {
    for (i = 0; i < function->n_arguments; i++) {
      if (function->var_argument[i] != NULL) { 
        wasora_var(function->var_argument[i]) = x[i];
      }
    }
  }
  
  return;
  
}

// evalua un token que es una funcion
// el chiste es que los argumentos estan dentro de la estructura
double wasora_evaluate_factor_function(factor_t *token) {

  int i;
  double *x;
  double x0;
  double y;


  wasora_push_error_message("when computing function %s", token->function->name);

  if (token->function->n_arguments == 1) {

    // para no tener que hacer malloc y frees todo el tiempo, para un solo argumento lo
    // hacemos diferente
    x0 = wasora_evaluate_expression(&token->arg[0]);
    y = wasora_evaluate_function(token->function, &x0);

  } else {

    // muchos argumentos necesitan vectores
    x = malloc(token->function->n_arguments*sizeof(double));

    for (i = 0; i < token->function->n_arguments; i++) {
      x[i] = wasora_evaluate_expression(&token->arg[i]);
    }

    y = wasora_evaluate_function(token->function, x);

    free(x);
    
  }

  if (gsl_isnan(y) || gsl_isinf(y)) {
    wasora_nan_error();
  }

  wasora_pop_error_message();  

  return y;

}
// evalua una function en el punto *x
double wasora_evaluate_function(function_t *function, const double *x) {

  int i;
  double y = 0;

  if (function->type == type_pointwise_mesh_node) {
    return wasora.interpolate_function_node(function, x);
    
  } else if (function->type == type_pointwise_mesh_cell) {
    return wasora.interpolate_function_cell(function, x);
    
  } else if (function->type == type_pointwise_mesh_property) {
    return wasora.interpolate_function_property(function, x);
    
  } else if (function->type == type_routine) {
    y = function->routine(x);
    
  } else if (function->algebraic_expression.n_tokens != 0 && function->n_arguments == 1) {

    // funcion con expresion algebraic unidimensional
    double x_old;

    x_old = wasora_value(function->var_argument[0]);
    wasora_value(function->var_argument[0]) = x[0];
    y = wasora_evaluate_expression(&function->algebraic_expression);
    wasora_value(function->var_argument[0]) = x_old;

  } else if (function->algebraic_expression.n_tokens != 0 && function->n_arguments > 1) {

    // funcion con expresion algebraic multidimensional
    double *vecx_old;

    vecx_old = malloc(function->n_arguments*sizeof(double));

    for (i = 0; i < function->n_arguments; i++) {
      vecx_old[i] = wasora_value(function->var_argument[i]);
      wasora_value(function->var_argument[i]) = x[i];
    }

    y = wasora_evaluate_expression(&function->algebraic_expression);

    for (i = 0; i < function->n_arguments; i++) {
      wasora_value(function->var_argument[i]) = vecx_old[i];
    }

    free(vecx_old);

  } else if (function->data_size != 0 && function->n_arguments == 1) {

    // funcion definida por puntos unidimensional
    // OJO! la GSL 1.15 no extrapola, devuelve nan y a comerla, asi que ante
    // de que nos devuelva porquerias miramos que onda
    if (x[0] < function->data_argument[0][0]) {
      y = function->data_value[0] - (function->data_value[1]-function->data_value[0])/(function->data_argument[0][1]-function->data_argument[0][0]) * (function->data_argument[0][0] - x[0]);
    } else if (x[0] > function->data_argument[0][function->data_size-1]) {
      y = function->data_value[function->data_size-1] + (function->data_value[function->data_size-1]-function->data_value[function->data_size-2])/(function->data_argument[0][function->data_size-1]-function->data_argument[0][function->data_size-2]) * (x[0] - function->data_argument[0][function->data_size-1]);
    } else {
      if (function->type == type_pointwise_vector) {
        if (gsl_interp_init(function->interp, function->data_argument[0], function->data_value, function->data_size) != GSL_SUCCESS) {
          wasora_runtime_error();
        }
      }
      y = gsl_interp_eval(function->interp, function->data_argument[0], function->data_value, x[0], function->interp_accel);
    }

  } else if (function->data_size != 0 && function->n_arguments > 1) {

    // funcion definida por puntos multidimensional
    int j;
 
    if (function->multidim_interp == nearest) {

      if (function->kd != NULL) {
        
        // k-dimensional tree
        y = *((double *)kd_res_item_data(kd_nearest(function->kd, x)));
        
      } else if (function->rectangular_mesh) {
        
        int actual_min_index;

        int *a;
        int *b;
        int *c;
        int prod;

        a = malloc(function->n_arguments*sizeof(int));
        b = malloc(function->n_arguments*sizeof(int));
        c = malloc(function->n_arguments*sizeof(int));

        for (i = 0; i < function->n_arguments; i++) {

          a[i] = 0;
          b[i] = function->rectangular_mesh_size[i];

          while ((b[i]-a[i]) > 1) {
            c[i] = 0.5*(a[i]+b[i]);

            if (function->rectangular_mesh_point[i][c[i]] > x[i]) {
              b[i] = c[i];
            } else {
              a[i] = c[i];
            }
          }

          c[i] = ceil(0.5*(a[i]+b[i]));

          if (fabs(function->rectangular_mesh_point[i][c[i]-1] - x[i]) < fabs(function->rectangular_mesh_point[i][c[i]] - x[i])) {
            c[i]--;
          }

        }

        // una vez que encontramos todos los c, vemos cual es el indice posta
        actual_min_index = wasora_structured_scalar_index(function->n_arguments, function->rectangular_mesh_size, c, function->x_increases_first);

        free(c);
        free(b);
        free(a);

        y = function->data_value[actual_min_index];
        
      }
      

    } else if (function->multidim_interp == rectangle) {

      // flag que indica si nos pidieron un punto de la definicion
      int flag;
      double *f;
      int *ctilde;
      double shape;
      
      // indices para la biseccion
      int *a;
      int *b;
      int *c;
      // coordenadas normalizadas a [-1:1]
      double *r;

      a = malloc(function->n_arguments*sizeof(int));
      b = malloc(function->n_arguments*sizeof(int));
      c = malloc(function->n_arguments*sizeof(int));
      r = malloc(function->n_arguments*sizeof(double));

      flag = 1;  // suponemos que nos pidieron un punto del problema
      for (i = 0; i < function->n_arguments; i++) {

        a[i] = 0;
        b[i] = function->rectangular_mesh_size[i];

        while ((b[i]-a[i]) > 1) {
          c[i] = floor(0.5*(a[i]+b[i]));

          if (function->rectangular_mesh_point[i][c[i]] > x[i]) {
            b[i] = c[i];
          } else {
            a[i] = c[i];
          }
        }

        c[i] = floor(0.5*(a[i]+b[i]));
        if (gsl_fcmp(function->rectangular_mesh_point[i][c[i]], x[i], function->multidim_threshold) != 0) {
          flag = 0;
        }

      }

      if (flag) {
        // nos pidieron un punto de la definicion
        y = function->data_value[wasora_structured_scalar_index(function->n_arguments, function->rectangular_mesh_size, c, function->x_increases_first)];
      } else {
        // tenemos que interpolar
        for (i = 0; i < function->n_arguments; i++) {
          // calculamos las coordenadas normalizadas en [-1:1], teniendo en cuenta
          // que capaz nos pidieron extrapolar
          if (x[i] < function->rectangular_mesh_point[i][0]) {
            r[i] = -1;
          } else if (c[i] == function->rectangular_mesh_size[i]-1) {
            c[i]--;
            r[i] = 1;
          } else {
            r[i] = -1 + 2*(x[i] - function->rectangular_mesh_point[i][c[i]])/(function->rectangular_mesh_point[i][c[i]+1] - function->rectangular_mesh_point[i][c[i]]);
          } 
         
        }

        y = 0;
        ctilde = malloc(function->n_arguments * sizeof(int));
        for (i = 0; i < (1<<function->n_arguments); i++)  {
          shape = 1;
          for (j = 0; j < function->n_arguments; j++) {
            // el desarrollo binario de i nos dice si hay que sumar uno o no al indice
            ctilde[j] = c[j] + ((i & (1<<j)) != 0);
            // y tambien como es el termino correspondiente a la funcion de forma
            shape *= ((i & (1<<j)) == 0) ? (1.0-r[j]) : (1.0+r[j]);
          }
          y += shape * function->data_value[wasora_structured_scalar_index(function->n_arguments, function->rectangular_mesh_size, ctilde, function->x_increases_first)];
        }
        
        y *= 1.0/(1<<function->n_arguments);
        free(ctilde);
        
      }

      free(r);
      free(c);
      free(b);
      free(a);


/*   
    } else if (function->multidim_interp == triangle) {

      double *dist;
      int *index;
      int far_away;
      double leg;

      size_t *closest;

      gsl_matrix *A;
      gsl_vector *coeff;
      gsl_vector *value;
      gsl_permutation *p;
      int s;

      int err;


      // calculamos las distancias de todos los puntos dato al punto pedido
      // el calloc las inicializa en cero
      n = 0;
      dist = calloc(function->data_size, sizeof(double));
      index = calloc(function->data_size, sizeof(int));
      for (i = 0; i < function->data_size; i++) {

        far_away = 0;

        for (j = 0; j < function->n_arguments; j++) {

          // leg es cateto en ingles segun google
          leg = fabs(function->data_argument[j][i]-x[j]);

          // si nos dieron una distancia caracteristica, entonces no hace falta
          // calcular todas las distancias, aquellos puntos que tengan al menos
          // una coordenada que este a mas tres o cuatro distancias del punto
          // pedido le ponemos dist = infinito

          if (function->multidim_characteristic_length != 0 && leg > function->multidim_characteristic_length) {
            far_away = 1;
            break;
          }

          // TODO: hay que dar una metrica para calcular la distancia
          // cuando los ejes tienen unidades diferentes!!!
          dist[n] += leg*leg;

        }

        if (!far_away) {
          // si dist[n] es muy chiquito, nos pidieron justo un punto de los
          // de la tabla asi que devolvemos ese y a comerla
          if (dist[n] < function->multidim_threshold) {
            free(dist);
            free(index);
            return function->data_value[i];
          }

          // si no esta lejos nos acordamos de i e incrementamos n
          index[n] = i;
          n++;
        }

      }


      // aca vamos a guardar los indices de los puntos dato mas cercanos, pero
      // como eventualmente los mas cercanos caen en un plano capaz que tenemos
      // que buscar bastantes mas que el minimo indispensable
      closest = calloc(n, sizeof(size_t));

      // para calcular los coeficientes del plano tenemos que resolver un problema
      // de algebra lineal
      A = gsl_matrix_alloc(function->n_arguments+1, function->n_arguments+1);
      coeff = gsl_vector_alloc(function->n_arguments+1);
      value = gsl_vector_alloc(function->n_arguments+1);
      p = gsl_permutation_alloc(function->n_arguments+1);

      // elegimos los tres mas cercanos
      gsl_sort_smallest_index(closest, function->n_arguments+1, dist, 1, n);

      for (i = 0; i < function->n_arguments+1; i++) {
        // nos fabricamos una matriz para calcular los coeficientes del plano definido
        // por estos tres (o n) puntos
        for (j = 0; j < function->n_arguments; j++) {
          gsl_matrix_set(A, i, j, function->data_argument[j][index[closest[i]]]);
        }

        gsl_matrix_set(A, i, function->n_arguments, 1);

      //
      //    A x = b
      //
      //        [ x1   y1   1 ]            [ z1 ]           [ a ]
      //    A = [ x2   y2   1 ]       b =  [ z2 ]      x =  [ b ]
      //        [ x3   y4   1 ]            [ z3 ]           [ c ]
      //
      // donde *1, *2 y *3 son los puntos mas cercanos al pedido
      //  el plano es ax + by + c - z = 0
      /  entonces z = ax + by + c

        // y el vector de terminos independientes para resolver el problema
        gsl_vector_set(value, i, function->data_value[index[closest[i]]]);
      }

        // solve!
        for (i = 0; i < 3; i++) {
          for (j = 0; j < 3; j++) {
            printf("%lf  ", gsl_matrix_get(A, i, j));
          }
          printf("\n");
        }
        printf("\n");
      err = gsl_linalg_LU_decomp(A, p, &s);
      if ((err = gsl_linalg_LU_solve(A, p, value, coeff)) != GSL_SUCCESS) {

        int k;

        // si no se pudo resolver el sistema, los tres puntos estan alineados
           (o los n puntos forman un hiperplano de dimension n-1 en lugar de n)

        // devolvemos el mas cercano
        //y = function->data_value[closest[0]];

        // o... tiramos el tercer punto y buscamos otro que de un plano valido
        // en general, la que no sirve es la fila function->n_arguments (empezando
        // a contar de cero) de la matriz A
        // para eso podemos
        //   a. ordenar TODOS los puntos y vamos buscando alguno que de un plano valido
        //   b. vamos seleccionando los k mas cercanos y tomamos los dos primeros y el ultimo
        //      hasta que encontremos un plano valido
        //
        // opcion a: te la debo
        // opcion b
        k = function->n_arguments;

        do {
          k++;
          // elegimos los k+1 mas cercanos
          gsl_sort_smallest_index(closest, k+1, dist, 1, n);


          // tenemos que volver a escribir la matriz completa porque el LU la rompe
          for (i = 0; i < function->n_arguments; i++) {
            // arriba que quiere decir esta mantriz
            for (j = 0; j < function->n_arguments; j++) {
              gsl_matrix_set(A, i, j, function->data_argument[j][index[closest[i]]]);
            }
            gsl_matrix_set(A, i, function->n_arguments, 1);
            // y el vector de terminos independientes para resolver el problema
            gsl_vector_set(value, i, function->data_value[index[closest[i]]]);
          }

          // escribimos la ultima fila de la matriz con los nuevos valores
          for (j = 0; j < function->n_arguments; j++) {
            gsl_matrix_set(A, function->n_arguments, j, function->data_argument[j][index[closest[k]]]);
          }
          gsl_matrix_set(A, function->n_arguments, function->n_arguments, 1);

          // y el vector de terminos independientes para resolver el problema
          gsl_vector_set(value, function->n_arguments, function->data_value[index[closest[k]]]);

          // solve!
            for (i = 0; i < 3; i++) {
              for (j = 0; j < 3; j++) {
                printf("%lf  ", gsl_matrix_get(A, i, j));
              }
              printf("\n");
            }
          gsl_linalg_LU_decomp(A, p, &s);
          err = gsl_linalg_LU_solve(A, p, value, coeff);
        } while (err != GSL_SUCCESS);
      }


      // calculamos el punto a partir de la ecuacion del plano
      y = 0;

      for (i = 0; i < function->n_arguments; i++) {
        y += gsl_vector_get(coeff, i)*x[i];
      }
      y += gsl_vector_get(coeff, function->n_arguments);

      gsl_permutation_free(p);
      gsl_vector_free(value);
      gsl_vector_free(coeff);
      gsl_matrix_free(A);

      free(closest);

      free(dist);
      free(index);

*/
    }

  }


  if (gsl_isnan(y) || gsl_isinf(y)) {
    wasora_nan_error();
  }

  return y;

}

int wasora_structured_scalar_index(int n_dims, int *size, int *index, int x_increases_first) {
  int i, j;
  int k;
  int scalar_index = 0;
  
  if (x_increases_first) {
    for (i = 0; i < n_dims; i++) {
      k = 1;
      for (j = i-1; j >= 0; j--) {
        k *= size[j];
      }
      scalar_index += k*index[i];  
    }
  } else {
    for (i = n_dims-1; i >= 0; i--) {
      k = 1;
      for (j = i+1; j < n_dims; j++) {
        k *= size[j];
      }
      scalar_index += k*index[i];  
    }
  }
  
  return scalar_index;
  
}



double mesh_interpolate_function_node_dummy(function_t *function, const double *x) {
  return 0.0;
}

double mesh_interpolate_function_cell_dummy(function_t *function, const double *x) {
  return 0.0;
}

double mesh_interpolate_function_property_dummy(function_t *function, const double *x) {
  return 0.0;
}
