/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora initialization routines
 *
 *  Copyright (C) 2009--2013 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */

//#include "config.h"

#include <stdio.h>
#include <string.h>
#include <gsl/gsl_math.h>

#ifdef HAVE_READLINE
#include <readline/readline.h>
#endif

#include "wasora.h"
#include "thirdparty/kdtree.h"

// esta se llama despues de haber alocado los objetos
// la idea es definir variables, funciones, archivos, etc. internos
int wasora_init_before_parser(void) {

  int i;
  int error;

  // si nunca cargamos algun plugin que necesite mallas, todas las funciones
  // de este tipo se van a evaluar a cero
  // los plugins que tengan malla tienen que redireccionar estos apuntadores
  wasora.interpolate_function_node = mesh_interpolate_function_node_dummy;
  wasora.interpolate_function_cell = mesh_interpolate_function_cell_dummy;
  wasora.interpolate_function_property = mesh_interpolate_function_property_dummy;

//va+done_static+name done_static
//va+done_static+desc Flag that indicates whether the static calculation is over or not.
//va+done_static+desc It is set to true (i.e. $\neq 0$) by wasora if `step_static` $\ge$ `static_steps`.
//va+done_static+desc If the user sets it to true, the current step is marked as the last static step and
//va+done_static+desc the static calculation ends after finishing the step.
  wasora_special_var(done_static) = wasora_define_variable("done_static");
  

//va+done_transient+name done_transient
//va+done_transient+desc Flag that indicates whether the transient calculation is over or not.
//va+done_transient+desc It is set to true (i.e. $\neq 0$) by wasora if `t` $\ge$ `end_time`.
//va+done_transient+desc If the user sets it to true, the current step is marked as the last transient step and
//va+done_transient+desc the transient calculation ends after finishing the step.
  wasora_special_var(done_transient) = wasora_define_variable("done_transient");


//va+done_outer+name done_outer
//va+done_outer+desc Flag that indicates whether the parametric, optimization of fit calculation is over or not.
//va+done_outer+desc It is set to true (i.e. $\neq 0$) by wasora whenever the outer calculation is considered to be finished,
//va+done_outer+desc which can be that the parametric calculation swept the desidered parameter space or that the
//va+done_outer+desc optimization algorithm reached the desired convergence criteria.
//va+done_outer+desc If the user sets it to true, the current step is marked as the last outer step and
//va+done_outer+desc the transient calculation ends after finishing the step.
  wasora_special_var(done_outer) = wasora_define_variable("done_outer");


//va+done+name done
//va+done+desc Flag that indicates whether the overall calculation is over.
  wasora_special_var(done) = wasora_define_variable("done");

  

//va+step_static+name step_static
//va+step_static+desc Indicates the current step number of the iterative static calculation.
  wasora_special_var(step_static) = wasora_define_variable("step_static");
  

//va+step_transient+name step_transient
//va+step_transient+desc Indicates the current step number of the transient static calculation.
  wasora_special_var(step_transient) = wasora_define_variable("step_transient");
  

//va+step_outer+name step_outer
//va+step_outer+desc Indicates the current step number of the iterative outer calculation (parametric, optimization or fit).
  wasora_special_var(step_outer) = wasora_define_variable("step_outer");

//va+step_outer+name step_outer
//va+step_outer+desc Indicates the current step number of the iterative inner calculation (optimization or fit).
  wasora_special_var(step_inner) = wasora_define_variable("step_inner");

//va+in_outer_initial+name in_outer_initial
//va+in_outer_initial+desc Flag that indicates if the current step is the initial step of an optimization of fit run.
  wasora_special_var(in_outer_initial) = wasora_define_variable("in_outer_initial");
  

//va+in_static+name in_static
//va+in_static+desc Flag that indicates if wasora is solving the iterative static calculation.
  wasora_special_var(in_static) = wasora_define_variable("in_static");


//va+in_static+name in_static_first
//va+in_static+desc Flag that indicates if wasora is in the first step of the iterative static calculation.
  wasora_special_var(in_static_first) = wasora_define_variable("in_static_first");


//va+in_static+name in_static_last
//va+in_static+desc Flag that indicates if wasora is in the last step of the iterative static calculation.
  wasora_special_var(in_static_last) = wasora_define_variable("in_static_last");


//va+in_transient+name in_transient
//va+in_transient+desc Flag that indicates if wasora is solving transient calculation.
  wasora_special_var(in_transient) = wasora_define_variable("in_transient");
  

//va+in_transient_first+name in_transient_first
//va+in_transient_first+desc Flag that indicates if wasora is in the first step of the transient calculation.
  wasora_special_var(in_transient_first) = wasora_define_variable("in_transient_first");
  

//va+in_transient_last+name in_transient_last
//va+in_transient_last+desc Flag that indicates if wasora is in the last step of the transient calculation.
  wasora_special_var(in_transient_last) = wasora_define_variable("in_transient_last");


//va+static_steps+name static_steps
//va+static_steps+desc Number of steps that ought to be taken during the static calculation, to be set by the user. 
//va+static_steps+desc The default value is one, meaning only one static step. 
  wasora_special_var(static_steps) = wasora_define_variable("static_steps");
  wasora_var(wasora_special_var(static_steps)) = 1;
  

//va+end_time+name end_time
//va+end_time+desc Final time of the transient calculation, to be set by the user. 
//va+end_time+desc The default value is zero, meaning no transient calculation.
  wasora_special_var(end_time) = wasora_define_variable("end_time");
  

//va+t+name t
//va+t+desc Actual value of the time for transient calculations. This variable is set by
//va+t+desc wasora, but can be written by the user for example by importing it from another
//va+t+desc transient code by means of shared-memory objects. Care should be taken when
//va+t+desc solving DAE systems and overwriting `t`.
  wasora_special_var(t) = wasora_define_variable("t");
  wasora_special_var(time) = wasora_special_var(t);
  

//va+dt+name dt
//va+dt+desc Actual value of the time step for transient calculations. When solving DAE systems,
//va+dt+desc this variable is set by wasora. It can be written by the user for example by importing it from another
//va+dt+desc transient code by means of shared-memory objects. Care should be taken when
//va+dt+desc solving DAE systems and overwriting `t`. Default value is 1/16, which is
//va+dt+desc a power of two and roundoff errors are thus reduced.
  wasora_special_var(dt) = wasora_define_variable("dt");
  wasora_var(wasora_special_var(dt)) = 1.0/16.0;

//va+rel_error+name rel_error
//va+rel_error+desc Maximum allowed relative error for the solution of DAE systems. Default value is
//va+rel_error+desc is $1 \times 10^{-6}$. If a fine per-variable error control is needed, special vector
//va+rel_error+desc `abs_error` should be used.
  wasora_special_var(rel_error) = wasora_define_variable("rel_error");
  wasora_var(wasora_special_var(rel_error)) = 1e-6;
  

//va+min_dt+name min_dt
//va+min_dt+desc Minimum bound for the time step that wasora should take when solving DAE systems.
  wasora_special_var(min_dt) = wasora_define_variable("min_dt");
  

//va+max_dt+name max_dt
//va+max_dt+desc Maximum bound for the time step that wasora should take when solving DAE systems.
  wasora_special_var(max_dt) = wasora_define_variable("max_dt");
  

//va+i+name i
//va+i+desc Dummy index, used mainly in vector and matrix row subindex expressions.
  wasora_special_var(i) = wasora_define_variable("i");


//va+j+name j
//va+j+desc Dummy index, used mainly in matrix column subindex expressions.
  wasora_special_var(j) = wasora_define_variable("j");
  

//va+pi+name pi
//va+pi+desc A double-precision floating point representaion of the number $\pi$, equal to
//va+pi+desc `math.h` 's `M_PI` constant.
  wasora_special_var(pi) = wasora_define_variable("pi");
  wasora_var(wasora_special_var(pi)) = M_PI;
  

//va+zero+name zero
//va+zero+desc A very small positive number, which is taken to avoid roundoff 
//va+zero+desc errors when comparing floating point numbers such as replacing $a \leq a_\text{max}$
//va+zero+desc with $a < a_\text{max} +$ `zero`. Default is $(1/2)^{-50} \approx 9\times 10^{-16}$ .
  wasora_special_var(zero) = wasora_define_variable("zero");
  wasora_var(wasora_special_var(zero)) = ZERO;


//va+infinite+name infinite
//va+infinite+desc A very big positive number, which can be used as `end_time = infinite` or
//va+infinite+desc to define improper integrals with infinite limits. Default is $2^50 \approx 1 10^{15}$.
  wasora_special_var(infinite) = wasora_define_variable("infinite");
  wasora_var(wasora_special_var(infinite)) = INFTY;


//va+max_daughters+name max_daughters
//va+max_daughters+desc This variable should be set to the maximum number of daughter processes to fork when
//va+max_daughters+desc performing a parametric calculation in parallel. If it is equal to zero (or to one) then
//va+max_daughters+desc the computation is performed serially. If it is equal to minus one, then the number of online
//va+max_daughters+desc available cores is used. It is recommended to use a non-zero value only in special cases.
  wasora_special_var(max_daughters) = wasora_define_variable("max_daughters");


//va+on_nan+name on_nan
//va+on_nan+desc This should be set to a mask that indicates how to proceed if Not-A-Number signal (such as a division by zero)
//va+on_nan+desc is generated when evaluating any expression within wasora.
  wasora_special_var(on_nan) = wasora_define_variable("on_nan");

  
//va+on_gsl_error+name on_gsl_error
//va+on_gsl_error+desc This should be set to a mask that indicates how to proceed if an error ir raised in any
//va+on_gsl_error+desc routine of the GNU Scientific Library. 
  wasora_special_var(on_gsl_error) = wasora_define_variable("on_gsl_error");


//va+on_ida_error+name on_ida_error
//va+on_ida_error+desc This should be set to a mask that indicates how to proceed if an error ir raised in any
//va+on_ida_error+desc routine of the SUNDIALS IDA Library. 
  wasora_special_var(on_ida_error) = wasora_define_variable("on_ida_error");
  
  wasora_define_number("quit", 0);
  wasora_define_number("report", 0);
  wasora_define_number("dont_quit", ON_ERROR_NO_QUIT);
  wasora_define_number("dont_report", ON_ERROR_NO_REPORT);
  
//va+realtime_scale+name realtime_scale
//va+realtime_scale+desc If this variable is not zero, then the transient problem is run trying to syncrhonize the
//va+realtime_scale+desc problem time with realtime, up to a scale given. For example, if the scale is set to one, then
//va+realtime_scale+desc wasora will advance the problem time at the same pace that the real wall time advances. If set to
//va+realtime_scale+desc two, wasora's time wil advance twice as fast as real time, and so on. If the calculation time is
//va+realtime_scale+desc slower than real time modified by the scale, this variable has no effect on the overall behavior
//va+realtime_scale+desc and execution will proceed as quick as possible with no delays.
  wasora_special_var(realtime_scale) = wasora_define_variable("realtime_scale");

//   files 
  wasora.special_files.stdout_ = wasora_define_file("stdout", "stdout", 0, NULL, "w", 0);
  wasora.special_files.stdout_->pointer = stdout;

  wasora.special_files.stderr_ = wasora_define_file("stderr", "stderr", 0, NULL, "w", 0);
  wasora.special_files.stderr_->pointer = stderr;
  

  // e inicializamos los plugins
  for (i = 0; i < wasora.i_plugin; i++) {
    if ((error = wasora.plugin[i].init_before_parser()) != 0) {
      wasora_push_error_message("in plugin %s,", wasora.plugin[i].name_string);
      return error;
    }
  }

  return WASORA_PARSER_OK;

}

// esta se llama despues del parser, hay que verificar que everything este manzana
// y pedir shared memory objecst y esas cosas
int wasora_init_after_parser(void) {

  int i, j, k;
  int nx, ny, nz;
  function_t *function, *tmp_function;
  struct semaphore_t *semaphore;
  io_t *io;
  
  if (wasora.active_conditional_block != NULL) {
    wasora_push_error_message("conditional block not closed");
    return WASORA_RUNTIME_ERROR;
  }
  
  // primero inicializamos plugins porque por ahi inicializa cosas que se
  // usan en el common (como por ejemplo funciones sobre mallas estructuradas)
  for (i = 0; i < wasora.i_plugin; i++) {
    if (wasora.plugin[i].init_after_parser() != 0) {
      wasora_push_error_message("in plugin %s,", wasora.plugin[i].name_string);
      return WASORA_RUNTIME_ERROR;
    }
  }

  LL_FOREACH(wasora.semaphores, semaphore) {
    if ((semaphore->pointer = wasora_get_semaphore(semaphore->name)) == NULL) {
      wasora_push_error_message("opening semaphore '%s'", semaphore->name);
      return WASORA_RUNTIME_ERROR;
    }
  }  

  LL_FOREACH(wasora.ios, io) {
    // allocamos el objeto de memoria compartida 
    if (io->shm_name != NULL) {
/*      
      switch (io->floating_point_format) {
        case format_double:
*/
          if ((io->shm_pointer_double = wasora_get_shared_pointer(io->shm_name, io->size * sizeof(double))) == NULL) {
            wasora_push_error_message("\"%s\" allocating shared memory object '%s'", strerror(errno), io->shm_name);
            return WASORA_RUNTIME_ERROR;
          }
/*
        break;
        case format_float:
          if ((io->shm_pointer_float = wasora_get_shared_pointer(io->shm_name, io->size * sizeof(float))) == NULL) {
            wasora_push_error_message("\"%s\" allocating shared memory object '%s'", strerror(errno), io->shm_name);
            return WASORA_RUNTIME_ERROR;
          }
        break;
        case format_char:
          if ((io->shm_pointer_char = wasora_get_shared_pointer(io->shm_name, io->size * sizeof(char))) == NULL) {
            wasora_push_error_message("\"%s\" allocating shared memory object '%s'", strerror(errno), io->shm_name);
            return WASORA_RUNTIME_ERROR;
          }
        break;
      }
*/      
    }
  }
  
//   funciones 
  HASH_ITER(hh, wasora.functions, function, tmp_function) {

    if (function->data_size != 0) {
      if (function->n_arguments == 1) {
    
        if (function->interp_type.name == NULL) {
          // interpolacion 1D por default 
          function->interp_type = DEFAULT_INTERPOLATION;      
        }

        if ((function->interp = gsl_interp_alloc(&function->interp_type, function->data_size)) == NULL) {
          wasora_push_error_message("interpolation method for function '%s' needs more points", function->name);
          return WASORA_PARSER_ERROR;
        }

        function->interp_accel = gsl_interp_accel_alloc();

        // si no es de tipo vector, inicializamos el interpolador (los datos quedan fijos) 
        // porque si es vector y los vectores tienen fruta, el interpolador se queja 
        if (function->type != type_pointwise_vector) {
          wasora_push_error_message("in function %s", function->name);
          gsl_interp_init(function->interp, function->data_argument[0], function->data_value, function->data_size);
          wasora_pop_error_message();
        }
   
      } else {
    
        if (function->multidim_threshold == 0) {
          function->multidim_threshold = DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD;
        }

        if (function->n_arguments == 2) {

          // miramos si tiene una malla regular en 2d
          if (wasora_is_structured_grid_2d(function->data_argument[0], function->data_argument[1], function->data_size, &nx, &ny)) {
            function->rectangular_mesh = 1;
            function->x_increases_first = 1;
          } else if (wasora_is_structured_grid_2d(function->data_argument[1], function->data_argument[0], function->data_size, &ny, &nx)) {
            function->rectangular_mesh = 1;
            function->x_increases_first = 0;
          } else {
            function->rectangular_mesh = 0;
          }


          if (function->rectangular_mesh) {
            function->rectangular_mesh_size = malloc(2*sizeof(int));
            function->rectangular_mesh_size[0] = nx;
            function->rectangular_mesh_size[1] = ny;

            function->rectangular_mesh_point = malloc(2*sizeof(double *));
            function->rectangular_mesh_point[0] = malloc(nx*sizeof(double));
            function->rectangular_mesh_point[1] = malloc(ny*sizeof(double));

            if (function->x_increases_first) {
              for (j = 0; j < nx; j++) {
                function->rectangular_mesh_point[0][j] = function->data_argument[0][j];
              }
              for (j = 0; j < ny; j++) {
                function->rectangular_mesh_point[1][j] = function->data_argument[1][nx*j];
              }
            } else {
              for (j = 0; j < nx; j++) {
                function->rectangular_mesh_point[0][j] = function->data_argument[0][ny*j];
              }
              for (j = 0; j < ny; j++) {
                function->rectangular_mesh_point[1][j] = function->data_argument[1][j];
              }
            }

          }

        } else if (function->n_arguments == 3) {

          // miramos si tiene una malla regular en 3d
          if (wasora_is_structured_grid_3d(function->data_argument[0], function->data_argument[1], function->data_argument[2], function->data_size, &nx, &ny, &nz)) {
            function->rectangular_mesh = 1;
            function->x_increases_first = 1;
          } else if (wasora_is_structured_grid_3d(function->data_argument[2], function->data_argument[1], function->data_argument[0], function->data_size, &nz, &ny, &nx)) {
            function->rectangular_mesh = 1;
            function->x_increases_first = 0;
          } else {
            function->rectangular_mesh = 0;
          }


          if (function->rectangular_mesh) {
            function->rectangular_mesh_size = malloc(3*sizeof(int));
            function->rectangular_mesh_size[0] = nx;
            function->rectangular_mesh_size[1] = ny;
            function->rectangular_mesh_size[2] = nz;

            function->rectangular_mesh_point = malloc(3*sizeof(double *));
            function->rectangular_mesh_point[0] = malloc(nx*sizeof(double));
            function->rectangular_mesh_point[1] = malloc(ny*sizeof(double));
            function->rectangular_mesh_point[2] = malloc(nz*sizeof(double));

            if (function->x_increases_first) {
              for (j = 0; j < nx; j++) {
                function->rectangular_mesh_point[0][j] = function->data_argument[0][j];
              }
              for (j = 0; j < ny; j++) {
                function->rectangular_mesh_point[1][j] = function->data_argument[1][nx*j];
              }
              for (j = 0; j < nz; j++) {
                function->rectangular_mesh_point[2][j] = function->data_argument[2][nx*ny*j];
              }
            } else {
              for (j = 0; j < nx; j++) {
                function->rectangular_mesh_point[0][j] = function->data_argument[0][nz*ny*j];
              }
              for (j = 0; j < ny; j++) {
                function->rectangular_mesh_point[1][j] = function->data_argument[1][nz*j];
              }
              for (j = 0; j < nz; j++) {
                function->rectangular_mesh_point[2][j] = function->data_argument[2][j];
              }
            }

          }
        }
      }
    
      if (function->rectangular_mesh == 0 && function->multidim_interp == rectangle) {
        wasora_push_error_message("rectangular interpolation of function '%s' needs a rectangular mesh", function->name);
        return -1;
      }
    
      if (function->n_arguments > 3 && function->multidim_interp == rectangle) {
        wasora_push_error_message("rectangular interpolation does not work for dimensions higher than three (function '%s' has %d arguments)", function->name, function->n_arguments);
        return -1;
      }

      if (function->n_arguments > 1 && function->rectangular_mesh == 0 && function->multidim_interp == nearest ) {
          
        double *point;
          
        point = malloc(function->n_arguments*sizeof(double));
        function->kd = kd_create(function->n_arguments);
          
        for (j = 0; j < function->data_size; j++) {
          for (k = 0; k < function->n_arguments; k++) {
            point[k] = function->data_argument[k][j];
          }
          kd_insert(function->kd, point, &function->data_value[j]);
        }
          
        free(point);
      }
    }
  }

#ifdef HAVE_IDA
  if (wasora_dae.dimension > wasora.i_dae) {
    wasora_push_error_message("dimension of phase space is %d but %d equations were given", wasora_dae.dimension, wasora.i_dae);
    return WASORA_RUNTIME_ERROR;
  }
#endif
  
  
#ifdef HAVE_READLINE
  // autocompletion
  rl_attempted_completion_function = wasora_rl_completion;
#endif
  
  return WASORA_RUNTIME_OK;
}


int wasora_init_dae_initialize(void) {
  
#ifdef HAVE_IDA
  int i;
  int err;
  phase_object_t *phase_object;

  wasora_dae.x = N_VNew_Serial(wasora_dae.dimension);
  wasora_dae.dxdt = N_VNew_Serial(wasora_dae.dimension);
  wasora_dae.id = N_VNew_Serial(wasora_dae.dimension);

  wasora_dae.system = IDACreate();
  
  // llenamos el vector de ida con las variables del usuario 
  for (i = 0; i < wasora_dae.dimension; i++) {
    NV_DATA_S(wasora_dae.x)[i] = *(wasora_dae.phase_value[i]);
    NV_DATA_S(wasora_dae.dxdt)[i] = *(wasora_dae.phase_derivative[i]);
    // suponemos que son algebraicas, despues marcamos las diferenciales
    // pero hay que inicializar el vector id!
    NV_DATA_S(wasora_dae.id)[i] = DAE_ALGEBRAIC; 
  }
  // inicializamos IDA  
  if (IDAInit(wasora_dae.system, wasora_ida_dae, wasora_var(wasora_special_var(time)), wasora_dae.x, wasora_dae.dxdt) != IDA_SUCCESS) {
    return WASORA_RUNTIME_ERROR;
  }

  // seteo de tolerancias 
  if ((wasora.special_vectors.abs_error = wasora_get_vector_ptr("abs_error")) != NULL) {

    // si hay un vector de absolutas, lo usamos 
    if (wasora.special_vectors.abs_error->size != wasora_dae.dimension) {
      wasora_push_error_message("dimension of system is %d and dimension of abs_error is %d", wasora_dae.dimension, wasora.special_vectors.abs_error->size);
      return WASORA_RUNTIME_ERROR;
    }

    wasora_dae.abs_error = N_VNew_Serial(wasora_dae.dimension);

    for (i = 0; i < wasora_dae.dimension; i++) {
      if (gsl_vector_get(wasora_value_ptr(wasora.special_vectors.abs_error), i) > 0) {
        NV_DATA_S(wasora_dae.abs_error)[i] = gsl_vector_get(wasora_value_ptr(wasora.special_vectors.abs_error), i);
      } else {
        NV_DATA_S(wasora_dae.abs_error)[i] = wasora_var(wasora_special_var(rel_error));
      }
    }

    if (IDASVtolerances(wasora_dae.system, wasora_var(wasora_special_var(rel_error)), wasora_dae.abs_error) != IDA_SUCCESS) {
      return WASORA_RUNTIME_ERROR;
    }
  } else {
    // si no nos dieron vector de absolutas, ponemos absolutas igual a relativas 
    if (IDASStolerances(wasora_dae.system, wasora_var(wasora_special_var(rel_error)), wasora_var(wasora_special_var(rel_error))) != IDA_SUCCESS) {
      return WASORA_RUNTIME_ERROR;
    }
  }

  //  estas tienen que venir despues de IDAInit

  // marcamos como differential aquellos elementos del phase space que lo sean
  LL_FOREACH(wasora_dae.phase_objects, phase_object) {
    if (phase_object->differential) {
      for (i = phase_object->offset; i < phase_object->offset+phase_object->size; i++) {
        NV_DATA_S(wasora_dae.id)[i] = DAE_DIFFERENTIAL;
      }
    }
  }
        
  
  if (IDASetId(wasora_dae.system, wasora_dae.id) != IDA_SUCCESS) {
    return WASORA_RUNTIME_ERROR;
  }
 
  if (IDADense(wasora_dae.system, wasora_dae.dimension) != IDA_SUCCESS) {
    return WASORA_RUNTIME_ERROR;
  }
  if (IDASetInitStep(wasora_dae.system, wasora_var(wasora_special_var(dt))) != IDA_SUCCESS) {
    return WASORA_RUNTIME_ERROR;
  }
  if (wasora_var(wasora_special_var(max_dt)) != 0) {
    if (IDASetMaxStep(wasora_dae.system, wasora_var(wasora_special_var(max_dt))) != IDA_SUCCESS) {
      return WASORA_RUNTIME_ERROR;
    }
  }

  if (wasora_dae.initial_conditions_mode == from_variables) {
    if ((err = IDACalcIC(wasora_dae.system, IDA_YA_YDP_INIT, wasora_var(wasora_special_var(dt)))) != IDA_SUCCESS) {
      wasora_push_error_message("error computing initial conditions from variable values, error = %d", err);
      return WASORA_RUNTIME_ERROR;
    }
  } else if (wasora_dae.initial_conditions_mode == from_derivatives) {
    if ((err = IDACalcIC(wasora_dae.system, IDA_Y_INIT, wasora_var(wasora_special_var(dt)))) != IDA_SUCCESS)  {
      wasora_push_error_message("error computing initial conditions from derivative values, error = %d", err);
      return WASORA_RUNTIME_ERROR;
    }
  }  
  
  return WASORA_RUNTIME_OK;
#else

  wasora_push_error_message("wasora cannot solve DAE systems as it was not linked against SUNDIALS IDA library.");
  return WASORA_RUNTIME_ERROR;

#endif
  
}


// esta se ejecuta antes de cada corrida, hay que resetear el tiempo, borrar historias, etc
int wasora_init_before_run(void) {
  int i;
  history_t *history;
  print_t *print;
  file_t *file;
  
  wasora_var(wasora_special_var(step_static)) = 0;
  wasora_var(wasora_special_var(step_transient)) = 0;
  wasora_var(wasora_special_var(time)) = 0;
  wasora_var(wasora_special_var(dt)) = 1.0/16.0;
  wasora.current_time_path = &wasora.time_path[0];
  
  wasora_var(wasora_special_var(done_static)) = 0;
  wasora_var(wasora_special_var(done_transient)) = 0;
  wasora_var(wasora_special_var(done)) = 0;
  
#ifdef HAVE_IDA
  // el cuento es asi: cuando hacemos estudios parametricos, las condiciones
  // iniciales se vuelven a calcular porque ponemos t = 0
  // pero las derivadas valen lo ultimo que valian al terminar el paso anterior
  // asi que a menos que querramos hacer estudios adiabaticos (TODO!)
  // volvemos a poner las derivadas en cero
  // si el usuario da explicitamente los valores de las derivas, es lo mismo
  // que las condiciones iniciales asi que todo manzana
  if (wasora_dae.phase_derivative != NULL) {
    for (i = 0; i < wasora_dae.dimension; i++) {
      *(wasora_dae.phase_derivative[i]) = 0;
    }
  }
  
  if (wasora_dae.system != NULL) {
    IDAFree(&wasora_dae.system);
    wasora_dae.system = NULL;
  }
    
  if (wasora_dae.x != NULL) {
    N_VDestroy_Serial(wasora_dae.x);
    wasora_dae.x = NULL;
  }
  if (wasora_dae.dxdt != NULL) {
    N_VDestroy_Serial(wasora_dae.dxdt);
    wasora_dae.dxdt = NULL;
  }
  if (wasora_dae.id != NULL) {
    N_VDestroy_Serial(wasora_dae.id);
    wasora_dae.id = NULL;
  }  
#endif  
  
  LL_FOREACH(wasora.histories, history) {
    history->function->data_size = 0;
    history->position = 0;

    if (history->function->data_value != NULL) {
      free(history->function->data_value);
    }
    if (history->function->data_argument != NULL) {
      free(history->function->data_argument[0]);
      free(history->function->data_argument);
    }
  }  

  LL_FOREACH(wasora.prints, print) {
    print->last_static_step = 0;
    print->last_step = 0;
    print->last_time = 0;
    print->last_header_step = 0;
    print->header_already_printed = 0;
  }  

  for (file = wasora.files; file != NULL; file = file->hh.next) {
    if (file->pointer != NULL && fileno(file->pointer) > 2) {
      wasora_instruction_close_file(file);
    }
  }  

  for (i = 0; i < wasora.i_plugin; i++) {
    wasora_call(wasora.plugin[i].init_before_run());
  }
  
  return WASORA_RUNTIME_OK;
}



// mira si los n puntos en x[] y en y[] forman una grilla estruturada
int wasora_is_structured_grid_2d(double *x, double *y, int n, int *nx, int *ny) {

  int i, j;
    
  // hacemos una primera pasada sobre x hasta encontrar de nuevo el primer valor
  i = 0;
  do {
    if (++i == n) {   
      // si recorrimos todo el set y nunca aparecio otra vez el x[0] a comerla
      return 0;
    }
  } while (gsl_fcmp(x[i], x[0], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD) != 0);
  
  // si i es uno entonces a comerla
  if (i == 1) {
    return 0;
  }
    
  // asumimos que la cantidad de datos en x es i
  *nx = i;
  // y entonces ny tiene que ser la cantidad de datos totales dividido data_size
  if (n % (*nx) != 0) {
    // si esto no es entero, no puede ser rectangular
    return 0;
  }
  *ny = n/(*nx);

  // si x[j] != x[j-nx] entonces no es una grilla estructurada
  for (i = *nx; i < n; i++) {
    if (gsl_fcmp(x[i], x[i-(*nx)], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD*fabs(x[(*nx)-1]-x[0])) != 0) {
      return 0;
    }
  }
  
  // ahora miramos el conjunto y, tienen que ser nx valores igualitos consecutivos
  for (i = 0; i < *ny; i++) {
    for (j = 0; j < (*nx)-1; j++) {
      if (gsl_fcmp(y[i*(*nx) + j], y[i*(*nx) + j + 1], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD*fabs(y[n-1]-y[0])) != 0) {
        return 0;
      }
    }
  }
  

  // si llegamos hasta aca, es una grilla bidimensional estructurada!
  return 1;
    
}


// mira si los n puntos en x[], y[] y z[] forman una grilla estruturada
int wasora_is_structured_grid_3d(double *x, double *y, double *z, int n, int *nx, int *ny, int *nz) {

  int i;
  
  // hacemos una pasada sobre x hasta encontrar de nuevo el primer valor
  i = 0;
  do {
    i += 1;
    if (i == n) {   
      // si recorrimos todo el set y nunca aparecio otra vez el x[0] a comerla
      return 0;
    }
  } while (gsl_fcmp(x[i], x[0], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD) != 0);

  // si i es uno entonces a comerla
  if (i == 1) {
    return 0;
  }
    
  // asumimos que la cantidad de datos en x es i
  *nx = i;

  // hacemos una pasada sobre y hasta encontrar de nuevo el primer valor
  // pero con step nx
  i = 0;
  do {
    i += *nx;
    if (i == n) {   
      // si recorrimos todo el set y nunca aparecio otra vez el y[0] a comerla
      return 0;
    }
  } while (gsl_fcmp(y[i], y[0], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD) != 0);
  
  // si i es uno entonces a comerla
  if (i == *nx || i % ((*nx)) != 0) {
    return 0;
  }

  // asumimos que la cantidad de datos en x es i
  *ny = i/(*nx);
  
  
  // y entonces nz tiene que ser la cantidad de datos totales dividido data_size
  if (n % ((*ny)*(*nx)) != 0) {
    // si esto no es entero, no puede ser rectangular
    return 0;
  }
  *nz = n/((*ny)*(*nx));


  for (i = *nx; i < n; i++) {
    if (gsl_fcmp(x[i], x[i-(*nx)], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD*fabs(x[(*nx)-1]-x[0])) != 0) {
      return 0;
    }
  }
  for (i = (*nx)*(*ny); i < n; i++) {
    if (gsl_fcmp(y[i], y[i-(*nx)*(*ny)], DEFAULT_MULTIDIM_INTERPOLATION_THRESHOLD*fabs(x[(*nx)-1]-x[0])) != 0) {
      return 0;
    }
  }
  
  
  return 1;    
}


// mira si los n puntos en x[], y[] y z[] forman una grilla estruturada
/*
int wasora_is_structured_grid_xlast(int n_args, int n_points, double *arg, int *size) {
 
  int i, j;
  
  // checkeo rapido:
  if ((n_points % n_args) != 0) {
    return 0;
  }
  
  // primer paso, barremos para ver si dan los tamanios
  for (i = 0; i < n_args; i++) {
    
    // miramos cuando es la primera vez que cambia el argumento
    if (j >= n_points-1) {
      return 0;
    }
    size[i] = 1;
    j = 0;
    while (gsl_fcmp(arg[j], arg[j+1])
  
  
}
*/
