/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora parametric runs
 *
 *  Copyright (C) 2009--2014 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */


#include <string.h>
#include <unistd.h>
#include <math.h>

#ifndef __WIN32__
 #include <sys/types.h>
 #include <sys/wait.h>
#endif

#include <time.h>

#include <gsl/gsl_math.h>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_qrng.h>

#include "wasora.h"

void wasora_parametric_run(void) {
  int i;
  int step = 0;
  int *local_step;
  
  gsl_rng *r = NULL;
  gsl_qrng *q = NULL;
  double *v;
          
  
  if ((int)wasora_value(wasora_special_var(max_daughters)) == -1) {
    wasora_parametric_run_parallel();
  } else {

    local_step = calloc(wasora.parametric.dimensions, sizeof(int));
    v = calloc(wasora.parametric.dimensions, sizeof(double));
    wasora.parametric.nsteps = calloc(wasora.parametric.dimensions, sizeof(int));

    if (wasora.parametric.outer_steps == 0) {
      wasora.parametric.outer_steps = 1;
      for (i = 0; i < wasora.parametric.dimensions; i++) {
        wasora.parametric.nsteps[i] = (int)((wasora.parametric.max[i] - wasora.parametric.min[i])/wasora.parametric.step[i]);
        wasora.parametric.outer_steps *= 1 + wasora.parametric.nsteps[i];
      }
    } else {
      for (i = 0; i < wasora.parametric.dimensions; i++) {
        wasora.parametric.step[i] = (wasora.parametric.max[i] - wasora.parametric.min[i])/pow((double)wasora.parametric.outer_steps, 1.0/(double)wasora.parametric.dimensions);
        wasora.parametric.nsteps[i] = (int)((wasora.parametric.max[i] - wasora.parametric.min[i])/wasora.parametric.step[i]);
      }
    }
    
    // the offset only has meaning for rng methods
    wasora.parametric.outer_steps -= wasora.parametric.offset;
   
    for (step = 0; step < wasora.parametric.outer_steps; step++) {
      wasora_var(wasora.special_vars.step_outer) = (double)(step+1);

      // calculamos el vector de parametros
      switch (wasora.parametric.type) {
        case parametric_linear:
          for (i = 0; i < wasora.parametric.dimensions; i++) {
            v[i] = (double)local_step[i] * wasora.parametric.step[i]/(wasora.parametric.max[i] - wasora.parametric.min[i]);
          }
        break;
        case parametric_logarithmic:
          for (i = 0; i < wasora.parametric.dimensions; i++) {
//            v[i] = (gsl_sf_exp((double)local_step[i]/(double)(wasora.parametric.steps[i]-1))-1)/(M_E-1);
            v[i] = (wasora.parametric.min[i]*gsl_sf_exp((double)local_step[i]*gsl_sf_log(wasora.parametric.max[i]/wasora.parametric.min[i])/(double)(wasora.parametric.nsteps[i]-1))-wasora.parametric.min[i])/(wasora.parametric.max[i]-wasora.parametric.min[i]);
          }
          break;
        case parametric_random:
          if (r == NULL) {
            r = gsl_rng_alloc(DEFAULT_RANDOM_METHOD);
            gsl_rng_set(r, (unsigned long int)(time(NULL)));
          }
          for (i = 0; i < wasora.parametric.dimensions; i++) {
            v[i] = gsl_rng_uniform(r);
          }
        break;
        case parametric_gaussianrandom:
          if (r == NULL) {
            r = gsl_rng_alloc(DEFAULT_RANDOM_METHOD);
            gsl_rng_set(r, (unsigned long int)(time(NULL)));
          }
          for (i = 0; i < wasora.parametric.dimensions; i++) {
            // el rango es tres sigma
            v[i] = 0.5 + gsl_ran_gaussian(r, 0.5/3.0);
          }
        break;
        case parametric_sobol:
          if (q == NULL) {
            q = gsl_qrng_alloc(gsl_qrng_sobol, wasora.parametric.dimensions);
            for (i = 0; i < wasora.parametric.offset; i++) {
              gsl_qrng_get(q, v);
            }
          }
          gsl_qrng_get(q, v);
        break;
        case parametric_niederreiter:
          if (q == NULL) {
            q = gsl_qrng_alloc(gsl_qrng_niederreiter_2, wasora.parametric.dimensions);
            for (i = 0; i < wasora.parametric.offset; i++) {
              gsl_qrng_get(q, v);
            }
          }
          gsl_qrng_get(q, v);
        break;
        case parametric_halton:
          if (q == NULL) {
            q = gsl_qrng_alloc(gsl_qrng_halton, wasora.parametric.dimensions);
            for (i = 0; i < wasora.parametric.offset; i++) {
              gsl_qrng_get(q, v);
            }
          }
          gsl_qrng_get(q, v);
        break;
        case parametric_reversehalton:
          if (q == NULL) {
            q = gsl_qrng_alloc(gsl_qrng_reversehalton, wasora.parametric.dimensions);
            for (i = 0; i < wasora.parametric.offset; i++) {
              gsl_qrng_get(q, v);
            }
          }
          gsl_qrng_get(q, v);
        break;
      }
      
      for (i = 0; i < wasora.parametric.dimensions; i++) {
        wasora_value(wasora.parametric.variable[i]) = wasora.parametric.min[i] + v[i]*(wasora.parametric.max[i] - wasora.parametric.min[i]);
      }
      

      // miramos si es el ultimo pase
      wasora_var(wasora_special_var(done_outer)) = (step == wasora.parametric.outer_steps-1);

      // corremos el programa en modo estandar
      wasora_standard_run();
      
      // incrementamos steps
      ++local_step[wasora.parametric.dimensions-1];
      for (i = wasora.parametric.dimensions-2; i >= 0; i--) {
        if (local_step[i+1] == (wasora.parametric.nsteps[i+1]+1)) {
          local_step[i+1] = 0;
          ++local_step[i];
        }
      }
      
    }

    if (r != NULL) {
      gsl_rng_free(r);
    }
    if (q != NULL) {
      gsl_qrng_free(q);
    }
    free(local_step);
  }
  

  return;
  
}

void wasora_parametric_run_parallel(void (*main_routine)(void)) {

/*
  int i, j, k;

  wasora.special_vars.max_daughters->current[0] = (double)sysconf(_SC_NPROCESSORS_ONLN);
  
  wasora.daughters = 1;
  for (i = 0; i < wasora.n_parametrics; i++) {
    
    // evaluamos las expresiones
    wasora.parametric[i].min.value = wasora_evaluate_expression(&wasora.parametric[i].min);
    wasora.parametric[i].max.value = wasora_evaluate_expression(&wasora.parametric[i].max);
    wasora.parametric[i].step.value = wasora_evaluate_expression(&wasora.parametric[i].step);
    
    if (wasora.parametric[i].type == parametric_linear) {
      if (wasora.parametric[i].step.value <= 0) {
        wasora_push_error_message("parametric step less or equal than zero");
        wasora_runtime_error();
      }
      wasora.parametric[i].steps = ((wasora.parametric[i].max.value-wasora.parametric[i].min.value)/wasora.parametric[i].step.value) + 1;
    } else if (wasora.parametric[i].type == parametric_log) {
      if (wasora.parametric[i].step.value <= 1) {
        wasora_push_error_message("logaritmic step less or equal than one");
        wasora_runtime_error();
      }
      wasora.parametric[i].steps = ceil((gsl_sf_log(wasora.parametric[i].max.value)-gsl_sf_log(wasora.parametric[i].min.value))/gsl_sf_log(wasora.parametric[i].step.value));
    }
    
    wasora.daughters *= wasora.parametric[i].steps;
  }
  
  // inicializamos un semaforo con el valor de la cantidad de procesos
  // en paralelo que queremos tener corriendo
  if ((wasora.fork_sem_name = malloc(MAX_STRING)) == NULL) {
    wasora_push_error_message("allocating memory");
    wasora_runtime_error();
  }
  sprintf(wasora.fork_sem_name, "wasora-fork-%d", getpid());
  wasora.fork_sem = wasora_get_semaphore(wasora.fork_sem_name);
  for (i = 0; i < (int)wasora.special_vars.max_daughters->current[0]; i++) {
    // lo subimos de a uno
    sem_post(wasora.fork_sem);
  }
  

  for (wasora.daughter_parametric_id = 0; wasora.daughter_parametric_id < wasora.daughters; wasora.daughter_parametric_id++) {

    // esperamos a que haya lugar
    sem_wait(wasora.fork_sem);
    // si ya largamos mas que max_daughters, hacemos un wait antes
    // de largar el que sigue
    if (wasora.daughter_parametric_id > (int)wasora.special_vars.max_daughters->current[0]) {
      wait(NULL);
    }

    wasora.daughter_parametric_pid = fork();
    if (wasora.daughter_parametric_pid == 0) {
      
      for (i = 0; i < wasora.n_parametrics; i++) {
        // cuentita magica para asignar los valores que corresponden a las variables parametricas
        k = 1;
        for (j = 0; j < i; j++) {
          k *= wasora.parametric[j].steps;
        }
        
        if (wasora.parametric[i].type == parametric_linear ) {
          wasora.parametric[i].variable->current[0] = wasora.parametric[i].min.value + ((int)(wasora.daughter_parametric_id/k) % wasora.parametric[i].steps)*wasora.parametric[i].step.value;
        } else if (wasora.parametric[i].type == parametric_log) {
          wasora.parametric[i].variable->current[0] = gsl_sf_exp(gsl_sf_log(wasora.parametric[i].min.value) + ((int)(wasora.daughter_parametric_id/k) % wasora.parametric[i].steps)*gsl_sf_log(wasora.parametric[i].step.value));
        }
      }

      
      main_routine();

      // avisamos que ya pueden largar otro hijo
      sem_post(wasora.fork_sem);
      
      return;

    }
  }

  // esperamos a que todos los hijos nos prendan el semaforo
//  if (wasora.daughters > wasora.max_daughters) {
    for (i = 0; i < (int)wasora.special_vars.max_daughters->current[0]; i++) {
      // lo esperamos de a uno y con los wrappers de shmem.c
      sem_wait(wasora.fork_sem);
      // y hacemos el join
      wait(NULL);
    }
//  }

  wasora_free_semaphore(wasora.fork_sem, wasora.fork_sem_name);
*/
  return;
}


