/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora parser
 *
 *  Copyright (C) 2009--2014 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#ifndef __WIN32__
 #include <unistd.h>
 #include <dlfcn.h>
#endif

#include <gsl/gsl_sf.h>
#include <gsl/gsl_interp.h>

#include "wasora.h"
#include "builtin.h"

typedef struct varlist_t varlist_t;
struct varlist_t {
  var_t *var;
  varlist_t *next;
};


// parsea el wasora_input file principal, haciendo llamadas a parse_wasora_input_file()
// que a su vez se puede llamar recursivamente por los INCLUDEs
int wasora_parse_main_input_file(char *filepath) {

  // miramos el tamanio de la pagina para usar un buffer de tamanio razonable
  // si despues necesitamos mas vamos pidiendo a medida que nos quedamos cortos,
  // pero no gastamos un fangote de memoria al pedo
#ifndef __WIN32__
  wasora.page_size = (size_t)sysconf(_SC_PAGESIZE);
#else
  wasora.page_size = 4096;
#endif
  wasora.actual_buffer_size = wasora.page_size-64;
  wasora.line = malloc(wasora.actual_buffer_size);

  // parseamos el wasora_input buscando plugins
  wasora_call(wasora_parse_input_for_plugins(filepath, 0, 0));

  // inicializamos
  wasora_call(wasora_init_before_parser());

  // parseamos el wasora_input principal (que a su vez puede volver a
  // llamarse recursivamente usando INCLUDEs)
  wasora_call(wasora_parse_input_file(filepath, 0, 0));

  // inicializamos cosas que se tienen que inicializar despues de haber leido el wasora_input
  wasora_call(wasora_init_after_parser());

  free(wasora.line);

  return WASORA_PARSER_OK;

}

// parsea un archivo de texto (se puede llamar recursivamente por los includes)
int wasora_parse_input_file(char *filepath, int from, int to) {

  FILE *wasora_input_file;
  char *line;

  int line_num, delta_line_num;
  int i, n;
  int understood;

  if (filepath == NULL) {
    return WASORA_PARSER_OK;
  }

  if ((wasora_input_file = wasora_fopen(filepath, "r")) == NULL) {
    if (strcmp(filepath, wasora.argv[wasora.optind]) == 0) {
      wasora_push_error_message("input file '%s' could not be opened: %s", filepath, strerror(errno));
    } else {
      wasora_push_error_message("included file '%s' could not be opened: % ", filepath, strerror(errno));
    }
    return WASORA_PARSER_ERROR;
  }

  // parseamos linea por linea
  line_num = 0;
  while ((delta_line_num = wasora_read_line(wasora_input_file)) != 0) {
    // delta_line_num negativo indica que la linea pide un argumento $n no provisto
    // eso puede pasar en una linea ignorada por un include from/to asi que no es
    // un error catastrofico
    line_num += abs(delta_line_num);
    
    if (wasora.line[0] != '\0' && ((from == 0 || line_num >= from) && ((to == 0) || line_num <= to))) {
      // a menos que nos metamos aca, que quiere decir que esa linea no se ignora asi
      // que ahora si a comerla
      if (delta_line_num < 0) {
        wasora_push_error_message("input file needs at least an argument in commandline");
        return WASORA_PARSER_ERROR;
      }
      
      line = strdup(wasora.line);
      if ((n = wasora_parse_line(line)) == WASORA_PARSER_ERROR) {
        wasora_push_error_message("%s: %d:", filepath, line_num);
        return WASORA_PARSER_ERROR;
      } else if (n == WASORA_PARSER_UNHANDLED) {
        understood = 0;
        for (i = 0; i < wasora.i_plugin; i++) {
          // recuperamos la linea completa porque los plugins pueden romper line
          strcpy(line, wasora.line);
          if ((n = wasora.plugin[i].parse_line(line)) == WASORA_PARSER_ERROR) {
            wasora_push_error_message("%s: %d:", filepath, line_num);
            return WASORA_PARSER_ERROR;
          }
          if (n == WASORA_PARSER_OK) {
            understood = 1;
          }
        }
        if (understood == 0) {
          if (strlen(wasora.line) > 32) {
            wasora.line[30] = '.';
            wasora.line[31] = '.';
            wasora.line[32] = '.';
            wasora.line[33] = '\0';
          }
          wasora_push_error_message("unknown syntax '%s'", wasora.line);
          wasora_push_error_message("%s: %d:", filepath, line_num);
          return WASORA_PARSER_ERROR;
        }
      }
      wasora_pop_error_message();
      free(line);
    }
  }

  fclose(wasora_input_file);
  
  // everything apple
  return WASORA_PARSER_OK;

}




// wasora_parser
int wasora_parse_line(char *line) {

  int i, j;
  char *token;
  char *dummy;

  if ((token = wasora_get_next_token(line)) != NULL) {

// ---------------------------------------------------------------------
///kw+INCLUDE+usage INCLUDE
///kw+INCLUDE+desc Include another wasora input file.
///kw+INCLUDE+detail Includes the input file located in the string *filepath* at the current location.
///kw+INCLUDE+detail The effect is the same as copying and pasting the contents of the included file
///kw+INCLUDE+detail at the location of the `INCLUDE` keyword. The path can be relative or absolute.
///kw+INCLUDE+example include1.was included2.was include2.was include3.was
    if ((strcasecmp(token, "INCLUDE") == 0)) {
      double xi;
      char *filepath;
      int from = 0;
      int to = 0;
    
///kw+INCLUDE+usage <filepath>
      if ((filepath = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected file path");
        return WASORA_PARSER_ERROR;
      }

      while ((token = wasora_get_next_token(NULL)) != NULL) {
///kw+INCLUDE+usage [ FROM <numexpr> ]
        if (strcasecmp(token, "FROM") == 0) {
          wasora_call(wasora_parser_expression_in_string(&xi));
          if ((from = (int)(xi)) <= 0) {
            wasora_push_error_message("expected a positive line number");
            return WASORA_PARSER_ERROR;
          }
///kw+INCLUDE+usage [ TO <numexpr> ]
        } else if (strcasecmp(token, "TO") == 0) {
          wasora_call(wasora_parser_expression_in_string(&xi));
          if ((to = (int)(xi)) <= 0) {
            wasora_push_error_message("expected a positive line number");
            return WASORA_PARSER_ERROR;
          }
        }
      }

      if (wasora_parse_input_file(filepath, from, to) != 0) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------
///kw+ABORT+usage ABORT
///kw+ABORT+desc Catastrophically abort the execution and quit wasora.
///kw+ABORT+detail Whenever the instruction `ABORT` is executed, wasora quits without closing files
///kw+ABORT+detail or unlocking shared memory objects. The objective of this instruction is, as
///kw+ABORT+detail illustrated in the examples, either to debug complex input files and check the
///kw+ABORT+detail values of certain variables or to conditionally abort the execution using `IF`
///kw+ABORT+detail clauses.
///kw+ABORT+example abort1.was abort2.was
    } else if (strcasecmp(token, "ABORT") == 0) {

      if (wasora_define_instruction(wasora_instruction_abort, NULL) == NULL) {
        return WASORA_PARSER_ERROR;
      }
      return WASORA_PARSER_OK;


// ---------------------------------------------------------------------
///kw+DEFAULT_ARGUMENT_VALUE+usage DEFAULT_ARGUMENT_VALUE
///kw+DEFAULT_ARGUMENT_VALUE+desc Give a default value for an optional commandline argument.
///kw+DEFAULT_ARGUMENT_VALUE+detail If a `$n` construction is found in the input file but the
///kw+DEFAULT_ARGUMENT_VALUE+detail commandline argument was not given, the default behavior is to
///kw+DEFAULT_ARGUMENT_VALUE+detail fail complaining that an extra argument has to be given in the
///kw+DEFAULT_ARGUMENT_VALUE+detail commandline. With this keyword, a default value can be assigned if
///kw+DEFAULT_ARGUMENT_VALUE+detail no argument is given, thus avoiding the failure and making the argument
///kw+DEFAULT_ARGUMENT_VALUE+detail optional.
    } else if (strcasecmp(token, "DEFAULT_ARGUMENT_VALUE") == 0) {

///kw+DEFAULT_ARGUMENT_VALUE+usage <num_expr> <string>
      
      return WASORA_PARSER_OK;
// ---------------------------------------------------------------------
///kw+IMPLICIT+usage IMPLICIT
///kw+IMPLICIT+desc Defines whether implicit declaration of variables is allowed or not.
///kw+IMPLICIT+detail By default, wasora allows variables (but not vectors nor matrices) to be
///kw+IMPLICIT+detail implicitly declared. To avoid introducing errors due to typos, explicit
///kw+IMPLICIT+detail declaration of variables can be forced by giving `IMPLICIT NONE`.
///kw+IMPLICIT+detail Whether implicit declaration is allowed or explicit declaration is required
///kw+IMPLICIT+detail depends on the last `IMPLICIT` keyword given, which by default is `ALLOWED`.
///kw+ABORT+example implicit.was
    } else if (strcasecmp(token, "IMPLICIT") == 0) {


///kw+IMPLICIT+usage { NONE | ALLOWED }
      char *keywords[] = {"NONE", "ALLOWED", ""};
      int values[] = {1, 0, 0};
      wasora_call(wasora_parser_keywords_ints(keywords, values, &wasora.implicit_none));

      return WASORA_PARSER_OK;

 // ---------------------------------------------------------------------
///kw+TIME_PATH+usage TIME_PATH
///kw+TIME_PATH+desc Forces transient problems to pass through specific instants of time.
    } else if ((strcasecmp(token, "TIME_PATH") == 0)) {

///kw+TIME_PATH+usage [ expr_1 ] [ expr_2 ] ... [ expr_n ]
      i = 0;
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        // el +2 es porque necesito +1 de por si y otro +1 para guardar el ultimo vacio
        wasora.time_path = realloc(wasora.time_path, (i+2)*sizeof(expr_t));
        if (wasora_parse_expression(token, &wasora.time_path[i]) != 0) {
          return WASORA_PARSER_ERROR;
        }
        i++;
      }

      // ponemos en vacio el ultimo
      wasora.time_path[i].n_tokens = 0;

      // y apuntamos el current al primero
      wasora.current_time_path = &wasora.time_path[0];

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+INITIAL_CONDITIONS_MODE+usage INITIAL_CONDITIONS_MODE
///kw+INITIAL_CONDITIONS_MODE+desc Defines how initial conditions of DAE problems are computed.
    } else if ((strcasecmp(token, "INITIAL_CONDITIONS_MODE") == 0)) {

///kw+INITIAL_CONDITIONS_MODE+usage { AS_PROVIDED | FROM_VARIABLES | FROM_DERIVATIVES }
      char *keywords[] = {"AS_PROVIDED", "FROM_VARIABLES", "FROM_DERIVATIVES", ""};
      int values[] = {as_provided, from_variables, from_derivatives, 0};
      wasora_call(wasora_parser_keywords_ints(keywords, values, (int *)&wasora_dae.initial_conditions_mode));

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+LOAD_PLUGIN+usage LOAD_PLUGIN
///kw+LOAD_PLUGIN+desc Loads a wasora plug-in from a dynamic shared object.      
    } else if (strcasecmp(token, "LOAD_PLUGIN") == 0) {

///kw+LOAD_PLUGIN+usage <filepath>
      // ya hicimos lo que habia que hacer en otro lado
      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+LOAD_ROUTINE+usage LOAD_ROUTINE
///kw+LOAD_ROUTINE+desc Loads one or more routines from a dynamic shared object.
    } else if (strcasecmp(token, "LOAD_ROUTINE") == 0) {

      void *library;
      char *path;

///kw+LOAD_ROUTINE+usage <filepath>
      if (wasora_parser_string(&path) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }
#ifndef __WIN32__      
      if ((library = wasora_dlopen(path, RTLD_NOW | RTLD_GLOBAL)) == NULL) {
        wasora_push_error_message("'%s' when opening library '%s'", dlerror(), path);
        return WASORA_PARSER_ERROR;
      }
#else
      wasora_push_error_message("dlopen");
      return WASORA_PARSER_ERROR;
      
#endif

///kw+LOAD_ROUTINE+usage <routine_1> [ routine_2 ... routine_n ]
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        if (wasora_define_loadable_routine(token, library) == NULL) {
          wasora_push_error_message("from library '%s'", path);
          return WASORA_PARSER_ERROR;
        }
      }

      free(path);

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+NUMBER+usage NUMBER
///kw+NUMBER+desc Defines a named number.
    } else if ((strcasecmp(token, "NUMBER") == 0)) {

///kw+NUMBER+usage <name>
///kw+NUMBER+usage [ = ]
///kw+NUMBER+usage <num_expr>
      // ya los procesamos en la primera vuelta
      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+VAR+usage VAR
///kw+VAR+desc Defines one or more scalar variables.
    } else if ((strcasecmp(token, "VAR") == 0)) {

///kw+VAR+usage [ name_1 ] [ name_2 ] ... [ name_n ]
      // los nombres de las variables uno atras de otro como punalada de manco
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        if (wasora_define_variable(token) == NULL) {
          return WASORA_PARSER_ERROR;
        }
      }

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+CONST+usage CONST
///kw+CONST+desc Marks a scalar variable, vector or matrix as a constant.
    } else if (strcasecmp(token, "CONST") == 0) {

      var_t *var;
      vector_t *vector;
      matrix_t *matrix;

///kw+CONST+usage [ name_1 ] [ name_2 ] ... [ name_n ]
      // los nombres de las variables uno atras de otro como punalada de manco
      while ((token = wasora_get_next_token(NULL)) != NULL) {

        // puede ser que sea una variable que ya este definida
        // antes y quieren que sea constante, asi que miramos si ya existe
        if ((var = wasora_get_variable_ptr(token)) != NULL) {
          var->constant = 1;
        } else if ((vector = wasora_get_vector_ptr(token)) != NULL) {
          vector->constant = 1;
        } else if ((matrix = wasora_get_matrix_ptr(token)) != NULL) {
          matrix->constant = 1;
        } else {
          // definimos una nueva variable
          if ((var = wasora_define_variable(token)) == NULL) {
            return WASORA_PARSER_ERROR;
          }
          var->constant = 1;
        }
      }

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------      
///kw+ALIAS+usage ALIAS
///kw+ALIAS+desc Defines a scalar alias of an already-defined symbol.

    } else if ((strcasecmp(token, "ALIAS") == 0)) {

      char *dummy_par;
      expr_t expr_row, expr_col;
      int row = 0;
      int col = 0;
      var_t *var;
      var_t *variable = NULL;
      vector_t *vector = NULL;
      matrix_t *matrix = NULL;

///kw+ALIAS+usage <name>
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected alias name");
        return WASORA_PARSER_ERROR;
      }
      if ((var = wasora_define_variable(token)) == NULL) {
        wasora_push_error_message("indentifier '%s' is already in use (alias name should come before the aliased object)", token);
        return WASORA_PARSER_ERROR;
      }

///kw+ALIAS+usage <symbol>
      // el nombre de lo que existe
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected existing object");
        return WASORA_PARSER_ERROR;
      }

      if ((dummy_par = strchr(token, '(')) != NULL) {
        *dummy_par = '\0';
      }

      if ((matrix = wasora_get_matrix_ptr(token)) == NULL) {
        if ((vector = wasora_get_vector_ptr(token)) == NULL) {
          if ((variable = wasora_get_variable_ptr(token)) == NULL) {
           wasora_push_error_message("unknown symbol '%s' (the existing object should come after the alias name)", token);
           return WASORA_PARSER_ERROR;
          }
        }
      }

      // si hay parentesis son subindices
      if (dummy_par != NULL) {
        if (matrix == NULL && vector == NULL) {
          wasora_push_error_message("symbol '%s' cannot be subscripted", token);
          return WASORA_PARSER_ERROR;
        }
        *dummy_par = '(';

        if (matrix != NULL) {
          if (wasora_parse_range(dummy_par, '(', ',', ')', &expr_row, &expr_col) != 0) {
            return WASORA_PARSER_ERROR;
          }
          row = (int)wasora_evaluate_expression(&expr_row) - 1;
          col = (int)wasora_evaluate_expression(&expr_col) - 1;
        } else if (vector != NULL) {
          if ((dummy = strrchr(dummy_par, ')')) == NULL) {
            wasora_push_error_message("unmatched parenthesis for '%s'", token);
            return WASORA_PARSER_ERROR;
          }
          *dummy = '\0';
          if (wasora_parse_expression(dummy_par+1, &expr_row) != 0) {
            return WASORA_PARSER_ERROR;
          }
          *dummy = ')';
          row = (int)wasora_evaluate_expression(&expr_row) - 1;
        }
      }

      if (matrix != NULL) {
        wasora_realloc_variable_ptr(var, gsl_matrix_ptr(wasora_value_ptr(matrix), row, col), 0);
        free(var->initial_static);
        var->initial_static = gsl_matrix_ptr(matrix->initial_static, row, col);
        free(var->initial_transient);
        var->initial_transient = gsl_matrix_ptr(matrix->initial_transient, row, col);
      } else if (vector != NULL) {
        wasora_realloc_variable_ptr(var, gsl_vector_ptr(wasora_value_ptr(vector), row), 0);
        free(var->initial_static);
        var->initial_static = gsl_vector_ptr(vector->initial_static, row);
        free(var->initial_transient);
        var->initial_transient = gsl_vector_ptr(vector->initial_transient, row);
      } else if (variable != NULL) {
        wasora_realloc_variable_ptr(var, wasora_value_ptr(variable), 0);
        free(var->initial_static);
        var->initial_static = variable->initial_static;
        free(var->initial_transient);
        var->initial_transient = variable->initial_transient;
      }

      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------
///kw+VECTOR+usage VECTOR
///kw+VECTOR+desc Defines a vector.
    } else if ((strcasecmp(token, "VECTOR") == 0)) {

      char *vectorname = NULL;
      double *vectordata = NULL;
      double xi;
      int vectorsize = 0;

///kw+VECTOR+usage <name>
      if (wasora_parser_string(&vectorname) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      while ((token = wasora_get_next_token(NULL)) != NULL) {

///kw+VECTOR+usage SIZE <num_expr>
        if (strcasecmp(token, "SIZE") == 0) {
          wasora_call(wasora_parser_expression_in_string(&xi));
          vectorsize = (int)(round(xi));

///kw+VECTOR+usage [ DATA <num_expr_1> <num_expr_2> ... <num_expr_n> ]
        } else if (strcasecmp(token, "DATA") == 0) {

          if (vectorsize == 0) {
            wasora_push_error_message("DATA given before SIZE");
            return WASORA_PARSER_ERROR;
          }

          vectordata = calloc(vectorsize, sizeof(double));
          for (i = 0; i < vectorsize; i++) {

            if ((token = wasora_get_next_token(NULL)) == NULL) {
              wasora_push_error_message("less elements given than vector size (%d vs %d)", i, vectorsize);
              return WASORA_PARSER_ERROR;
            }

            vectordata[i] = wasora_evaluate_expression_in_string(token);
          }

///kw+VECTOR+usage [ FUNCTION_DATA <function> ]
        } else if (strcasecmp(token, "FUNCTION_DATA") == 0) {
          
          function_t *function;

          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected function name");
            return WASORA_PARSER_ERROR;
          }

          if ((function = wasora_get_function_ptr(token)) == NULL) {
            wasora_push_error_message("unknown function '%s'", token);
            return WASORA_PARSER_ERROR;
          }
          if (function->data_size == 0) {
            wasora_push_error_message("function '%s' is not point-wise defined", token);
            return WASORA_PARSER_ERROR;
          }
          if (vectorsize == 0) {
            vectorsize = function->data_size;
          } else if (vectorsize != function->data_size) {
            wasora_push_error_message("function '%s' has %d points and vector '%s' has size %d ", token, function->data_size, vectorname, vectorsize);
            return WASORA_PARSER_ERROR;
          }
          
          vectordata = function->data_value;

        } else {

          if (vectordata != NULL) {
            wasora_push_error_message("more data ('%s') than vector size %d", token, vectorsize);
          } else {
            wasora_push_error_message("unknown keyword '%s'", token);
          }
          return WASORA_PARSER_ERROR;

        }

      }

      if (vectorsize <= 0) {
        wasora_push_error_message("expected positive vector size");
        return WASORA_PARSER_ERROR;
      }

      // listoooo
      if (wasora_define_vector(vectorname, vectorsize, vectordata) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      free(vectorname);
      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------
///kw+MATRIX+usage MATRIX
///kw+MATRIX+desc Defines a matrix.
    } else if ((strcasecmp(token, "MATRIX") == 0)) {

      char *matrixname = NULL;
      double **data = NULL;
      double xi = 0;
      int rows = 0;
      int cols = 0;

///kw+MATRIX+usage <name>
      if (wasora_parser_string(&matrixname) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      while ((token = wasora_get_next_token(NULL)) != NULL) {

///kw+MATRIX+usage ROWS <num_expr>
        if (strcasecmp(token, "ROWS") == 0) {
          wasora_call(wasora_parser_expression_in_string(&xi));
          rows = (int)xi;
///kw+MATRIX+usage COLS <num_expr>
        } else if (strcasecmp(token, "COLS") == 0) {
          wasora_call(wasora_parser_expression_in_string(&xi));
          cols = (int)xi;
        
///kw+MATRIX+usage [ DATA num_expr_1 num_expr_2 ... num_expr_n ]
        } else if (strcasecmp(token, "DATA") == 0) {

          if (rows == 0) {
            wasora_push_error_message("DATA given before ROWS and COLS");
            return WASORA_PARSER_ERROR;
          }

          data = calloc(rows, sizeof(double *));
          for (i = 0; i < rows; i++) {

            data[i] = calloc(cols, sizeof(double));
            for (j = 0; j < cols; j++) {

              if ((token = wasora_get_next_token(NULL)) == NULL) {
                wasora_push_error_message("less elements given than matrix size");
                return WASORA_PARSER_ERROR;
              }

              data[i][j] = wasora_evaluate_expression_in_string(token);
            }
          }
        } else {
          wasora_push_error_message("unknown keyword '%s'", token);
          return WASORA_PARSER_ERROR;
        }
      }

      if (wasora_define_matrix(matrixname, rows, cols, data) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      free(matrixname);
      return WASORA_PARSER_OK;

// ---------------------------------------------------------------------
///kw+FUNCTION+usage FUNCTION
///kw+FUNCTION+desc Defines a function of one or more variables.
    } else if (strcasecmp(token, "FUNCTION") == 0 || ((dummy = strchr(wasora.line, '=')) != NULL && *(dummy-1) == ':')) {

      function_t *function;
      char *argument, *dummy_argument;
      char *backup1, *backup2;
      char char_backup;
      int level, nargs;

///kw+FUNCTION+usage <name>(<var_1>[,var2,...,var_n])
      // si nos dieron el keyword FUNCTION tenemos que hacer un wasora_get_next_token
      // para obtener el nombre de la funcion, sino ya esta en token
      if (strcasecmp(token, "FUNCTION") == 0) {
        if ((token = wasora_get_next_token(NULL)) == NULL) {
          wasora_push_error_message("expected function name");
          return WASORA_PARSER_ERROR;
        }
      }



      // parseamos los argumentos
      if ((dummy = strchr(token, '(')) == NULL) {
        wasora_push_error_message("expected opening parenthesis '(' after function name '%s' (no spaces allowed)", token);
        return WASORA_PARSER_ERROR;
      }

      argument = strdup(dummy);
      wasora_strip_blanks(argument);

      if ((nargs = wasora_count_arguments(argument)) <= 0) {
        return WASORA_PARSER_ERROR;
      }

      dummy = strchr(token, '(');
      *dummy = '\0';

      if ((function = wasora_define_function(token, nargs)) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      function->arg_name = malloc(nargs * sizeof(char *));
      function->var_argument = malloc(nargs * sizeof(var_t *));
      function->column = malloc((nargs+1) * sizeof(int));
      // por default ponemos en las columnas 1 2 3 4 ...
      for (i = 0; i < nargs+1; i++) {
        function->column[i] = i+1;
      }



      if (nargs == 1) {
        // si la funcion toma un solo argumento es facil
        function->arg_name[0] = strdup(argument+1);
        if ((dummy = strchr(function->arg_name[0], ')')) == NULL) {
          wasora_push_error_message("expected ')' after function name (no spaces allowed)");
          return WASORA_PARSER_ERROR;
        }
        *dummy = '\0';

      } else {

        // si la funcion toma mas argumentos tenemos que buscarlos de a uno

        // si incrementamos argument despues no podemos hacer free, asi
        // que tenemos un apuntador dummy para argument
        dummy_argument = argument;

        for (i = 0; i < nargs; i++) {
          if (i != nargs-1) {

            // si no estamos en el ultimo argumento, tenemos que
            // poner un ")\0" donde termina el argumento
            level = 1;
            dummy = dummy_argument+1;
            while (1) {
              if (*dummy == ',' && level == 1) {
                break;
              }
              if (*dummy == '(') {
                level++;
              } else if (*dummy == ')') {
                level--;
              } else if (*dummy == '\0') {
                wasora_push_error_message("when parsing arguments");
                return WASORA_PARSER_ERROR;
              }
              dummy++;
            }

            // ponemos un \0 despues de dummy, pero necesitamos despues
            // restaurar el caracter que habia ahi
            char_backup = *(dummy);
            *(dummy) = '\0';
          } else {

            // le sacamos el ) final
            dummy[strlen(dummy)-1] = '\0';

          }


          // en dummy_argument+1 tenemos el argumento i-esimo
          function->arg_name[i] = strdup(dummy_argument+1);


          if (i != nargs-1) {
            *(dummy) = char_backup;
            dummy_argument = dummy;
          }
        }
      }
      free(argument);


      // linkeamos los argumentos a las variables correctas
      for (i = 0; i < nargs; i++) {
        if ((function->var_argument[i] = wasora_get_or_define_variable_ptr(function->arg_name[i])) == NULL) {
          return WASORA_PARSER_ERROR;
        }
      }


      while ((token = wasora_get_next_token(NULL)) != NULL && strcasecmp(token, "DATA") != 0) {
///kw+FUNCTION+usage {
///kw+FUNCTION+usage [ = <expr> ] |
        if (strcasecmp(token, "=") == 0 || strcasecmp(token, ":=") == 0) {

          function->type = type_algebraic;

          // todo lo que sigue al ':=' o al '=' deberia ser la expresion
          // no funciona el ()?:*/
//          dummy = token + (token[0] == ':')? 3 : 2;
          if (token[0] == ':') {
            dummy = token + 3;
          } else {
            dummy = token + 2;
          }

          if (wasora_parse_expression(dummy, &function->algebraic_expression) != 0) {
            return WASORA_PARSER_ERROR;
          }

          // terminamos con esta linea
          break;

///kw+FUNCTION+usage [ FILE_PATH <filepath> ] |
        } else if (strcasecmp(token, "FILE_PATH") == 0) {

          // TODO: el keyword "FILE" deberia querer decir otra cosa que "FILE_PATH"
          // TODO: poder pedir que se refresquen los datos en cada paso
          function->type = type_pointwise_file;
          wasora_parser_string(&function->data_file);

///kw+FUNCTION+usage [ ROUTINE <name> ] |
        } else if (strcasecmp(token, "ROUTINE") == 0) {

          function->type = type_routine;

          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected routine name");
            return WASORA_PARSER_ERROR;
          }

          if ((function->routine = wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("undefined routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+FUNCTION+usage [ VECTOR_DATA <vector_1> <vector_2> ... <vector_n> <vector_n+1> ]
///kw+FUNCTION+usage }
        } else if (strcasecmp(token, "VECTOR_DATA") == 0 || strcasecmp(token, "VECTORS") == 0) {

          function->type = type_pointwise_vector;
          function->vector_argument = calloc(nargs, sizeof(vector_t *));
          function->data_argument = calloc(nargs, sizeof(double *));
          function->data_size = 0;

          for (i = 0; i < nargs; i++) {
            if (wasora_parser_vector(&function->vector_argument[i]) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }

            if (function->data_size == 0) {
              function->data_size = function->vector_argument[i]->size;
            } else {
              if (function->vector_argument[i]->size != function->data_size) {
                wasora_push_error_message("vector sizes do not match (%d and %d)", function->data_size, function->vector_argument[i]->size);
                return WASORA_PARSER_ERROR;
              }
            }

            function->data_argument[i] = gsl_vector_ptr(wasora_value_ptr(function->vector_argument[i]), 0);
          }

          if (wasora_parser_vector(&function->vector_value) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

          if (function->vector_value->size != function->data_size) {
            wasora_push_error_message("vector sizes do not match (%d and %d)", function->data_size, function->vector_value->size);
            return WASORA_PARSER_ERROR;
          }

          function->data_value = gsl_vector_ptr(wasora_value_ptr(function->vector_value), 0);

///kw+FUNCTION+usage [COLUMNS <num_expr_1> <num_expr_2> ... <num_expr_n> <num_expr_n+1> ]
        } else if (strcasecmp(token, "COLUMNS") == 0) {

          if (nargs == 0) {
            wasora_push_error_message("number of arguments is equal to zero");
            return WASORA_PARSER_ERROR;
          }

          for (i = 0; i < nargs+1; i++) {
            if ((token = wasora_get_next_token(NULL)) == NULL) {
              wasora_push_error_message("expected %d columns specifications", nargs);
              return WASORA_PARSER_ERROR;
            }
            function->column[i] = (int)(wasora_evaluate_expression_in_string(token));

          }

///kw+FUNCTION+usage [ INTERPOLATION_THRESHOLD <expr> ]
        } else if (strcasecmp(token, "INTERPOLATION_THRESHOLD") == 0) {

           wasora_parser_expression_in_string(&function->multidim_threshold);

///kw+FUNCTION+usage [ INTERPOLATION
        } else if (strcasecmp(token, "INTERPOLATION") == 0) {

          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected interpolation method");
            return WASORA_PARSER_ERROR;
          }

///kw+FUNCTION+usage {
///kw+FUNCTION+usage linear |
          if (strcasecmp(token, "linear") == 0) {
            function->interp_type = *gsl_interp_linear;
///kw+FUNCTION+usage polynomial |
          } else if (strcasecmp(token, "polynomial") == 0) {
            function->interp_type = *gsl_interp_polynomial;
///kw+FUNCTION+usage spline |
          } else if (strcasecmp(token, "spline") == 0 || strcasecmp(token, "cspline") == 0 || strcasecmp(token, "splines") == 0) {
            function->interp_type = *gsl_interp_cspline;
///kw+FUNCTION+usage spline_periodic |
          } else if (strcasecmp(token, "spline_periodic") == 0 || strcasecmp(token, "cspline_periodic") == 0 || strcasecmp(token, "splines_periodic") == 0) {
            function->interp_type = *gsl_interp_cspline_periodic;
///kw+FUNCTION+usage akima |
          } else if (strcasecmp(token, "akima") == 0) {
            function->interp_type = *gsl_interp_akima;
///kw+FUNCTION+usage akima_periodic |
          } else if (strcasecmp(token, "akima_periodic") == 0) {
            function->interp_type = *gsl_interp_akima_periodic;
///kw+FUNCTION+usage nearest |
          } else if (strcasecmp(token, "nearest") == 0) {
            function->multidim_interp = nearest;
///kw+FUNCTION+usage rectangle
          } else if (strcasecmp(token, "rectangle") == 0 || strcasecmp(token, "rectangular") == 0) {
            function->multidim_interp = rectangle;
///kw+FUNCTION+usage } ]
          } else {
            wasora_push_error_message("undefined interpolation method '%s'", token);
            return WASORA_PARSER_ERROR;
          }

        } else {
          wasora_push_error_message("unknown keyword '%s'", token);
          return WASORA_PARSER_ERROR;
        }
      }


      // si no nos dieron expresion algebraica, hay que leer e interpolar datos
      if (function->algebraic_expression.n_tokens == 0) {
        if (function->type == type_pointwise_file) {

          // si nos dieron un archivo, leemos de ahi
          FILE *data_file;
          char data_line[MAX_STRING*MAX_STRING];
          int n_columns;

          if ((data_file = wasora_fopen(function->data_file, "r")) == NULL) {
            wasora_push_error_message("\"%s\" opening file '%s'", strerror(errno), function->data_file);
            return WASORA_PARSER_ERROR;
          }

          // contamos cuantas lineas no vacias hay -> ese es el tamanio del size
          function->data_size = 0;
          while (wasora_read_data_line(data_file, data_line) != 0) {
            if (data_line[0] != '\0') {
              // en la primera vuelta miramos cuantas columnas hay
              if (function->data_size == 0) {
                n_columns = 0;
                token = strtok(data_line, " \t");
                while (token != NULL) {
                  n_columns++;
                  token = strtok(NULL, " \t");
                }

                if (n_columns < (nargs+1)) {
                  wasora_push_error_message("at least %d columns expected but %d were given in  file '%s'", nargs+1, n_columns, function->data_file);
                  return WASORA_PARSER_ERROR;
                }

              }
              function->data_size++;
            }
          }

          function->data_argument = calloc(nargs, sizeof(double *));
          for (i = 0; i < nargs; i++) {
            function->data_argument[i] = calloc(function->data_size, sizeof(double));
          }
          function->data_value = calloc(function->data_size, sizeof(double));

          rewind(data_file);

          // ahora leemos los datos
          j = 0;
          while (wasora_read_data_line(data_file, data_line) != 0) {

            if (data_line[0] != '\0') {
              for (i = 0; i < nargs; i++) {
                if ((token = wasora_get_nth_token(data_line, function->column[i])) == NULL) {
                  wasora_push_error_message("wrong-formatted file '%s' at line %d", function->data_file, j+1);
                  return WASORA_PARSER_ERROR;
                }
                sscanf(token, "%lf", &function->data_argument[i][j]);
                free(token);

                // esto es para poder poner steps en funciones undimensionales
                if (nargs == 1 && j >= 2 && function->data_argument[i][j] == function->data_argument[i][j-1]) {
                  function->data_argument[i][j] += 0.005*(function->data_argument[i][j-1]-function->data_argument[i][j-2]);
                }


              }
              if ((token = wasora_get_nth_token(data_line, function->column[i])) == NULL) {
                wasora_push_error_message("not enough columns in file '%s' at line %d", function->data_file, j+1);
                return WASORA_PARSER_ERROR;
              }
              sscanf(token, "%lf", &function->data_value[j]);
              free(token);

              j++;
            }
          }

          fclose(data_file);


///kw+FUNCTION+usage [ DATA <num_expr_1> <num_expr_2> ... <num_expr_N> ]
        } else if (token != NULL && strcasecmp(token, "DATA") == 0) {
          // leemos del mismo archivo de entrada
          dummy = token + strlen(token)+1;

          // nos dieron los datos hard en el wasora_input
          backup1 = malloc(strlen(dummy)+8);
          sprintf(backup1, "dummy %s", dummy);
          backup2 = malloc(strlen(dummy)+8);
          sprintf(backup2, "dummy %s", dummy);

          // contamos cuanta informacion hay
          function->data_size = 0;
          if ((token = wasora_get_next_token(backup1)) == NULL) {
            return WASORA_PARSER_ERROR;
          }
          while ((token = wasora_get_next_token(NULL)) != NULL) {
            function->data_size++;
          }
          if (function->data_size % (nargs+1) != 0) {
            wasora_push_error_message("data mismatch for function '%s'", function->name);
            return WASORA_PARSER_ERROR;
          }
          function->data_size /= (nargs+1);

          function->data_argument = calloc(nargs, sizeof(double *));
          for (i = 0; i < nargs; i++) {
            function->data_argument[i] = calloc(function->data_size, sizeof(double));
          }
          function->data_value = calloc(function->data_size, sizeof(double));

          // leemos la informacion
          if ((token = wasora_get_next_token(backup2)) == NULL) {
            return WASORA_PARSER_ERROR;
          }
          for (i = 0; i < function->data_size; i++) {
            // argumentos
            for (j = 0; j < nargs; j++) {
              if ((token = wasora_get_next_token(NULL)) == NULL) {
                return WASORA_PARSER_ERROR;
              }
              function->data_argument[j][i] = wasora_evaluate_expression_in_string(token);

              //  para poder meter steps
              if (nargs == 1 && j >= 2 && function->data_argument[i][j] == function->data_argument[i][j-1]) {
                function->data_argument[i][j] += 0.005*(function->data_argument[i][j-1]-function->data_argument[i][j-2]);
              }
            }

            // valor
            if ((token = wasora_get_next_token(NULL)) == NULL) {
              return WASORA_PARSER_ERROR;
            }
            function->data_value[i] = wasora_evaluate_expression_in_string(token);
          }

          free(backup2);
          free(backup1);
        }


        // creamos tres variables (constantes) extra: func_a, func_b y func_n
        if (nargs == 1 && function->data_size != 0) {
          var_t *dummy_var;
          char *dummy_aux;

          dummy_aux = malloc(strlen(function->name) + 4);

          sprintf(dummy_aux, "%s_a", function->name);
          if ((dummy_var = wasora_define_variable(dummy_aux)) == NULL) {
            return WASORA_PARSER_ERROR;
          }
          wasora_realloc_variable_ptr(dummy_var, &function->data_argument[0][0], 0);

          sprintf(dummy_aux, "%s_b", function->name);
          if ((dummy_var = wasora_define_variable(dummy_aux)) == NULL) {
            return WASORA_PARSER_ERROR;
          }
          wasora_realloc_variable_ptr(dummy_var, &function->data_argument[0][function->data_size-1], 0);

          sprintf(dummy_aux, "%s_n", function->name);
          if ((dummy_var = wasora_define_variable(dummy_aux)) == NULL) {
            return WASORA_PARSER_ERROR;
          }
          wasora_value(dummy_var) = (double)function->data_size;

          free(dummy_aux);
        }

      }

      return WASORA_PARSER_OK;

// --- FILE -----------------------------------------------------
///kw+FILE+desc Defines a file, either as input or as output, for further reference.
///kw+FILE+usage < FILE | OUTPUT_FILE | INPUT_FILE >
    } else if (strcasecmp(token, "FILE") == 0 || strcasecmp(token, "OUTPUT_FILE") == 0 || strcasecmp(token, "INPUT_FILE") == 0) {

      file_t *file;
      char *name = NULL;
      char *format = NULL;
      char *mode = NULL;
      expr_t *arg = NULL;
      int n_args = 0;
      int open = 0;

      if (strcasecmp(token, "OUTPUT_FILE") == 0) {
        mode = strdup("w");
      } else if (strcasecmp(token, "INPUT_FILE") == 0) {
        mode = strdup("r");
      }

///kw+FILE+usage <name>
      if (wasora_parser_string(&name) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

///kw+FILE+usage <printf_format>
///kw+FILE+usage [ expr_1 expr_2 ... expr_n ]
      if (wasora_parser_string_format(&format, &n_args) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      arg = calloc(n_args, sizeof(expr_t));
      for (i = 0; i < n_args; i++) {
        if (wasora_parser_expression(&arg[i]) != WASORA_PARSER_OK) {
          return WASORA_PARSER_ERROR;
        }
      }

///kw+FILE+usage [ INPUT | OUTPUT | MODE <fopen_mode> ]
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        if (strcasecmp(token, "MODE") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a mode");
            return WASORA_PARSER_ERROR;
          }
          mode = strdup(token);
        } else if (strcasecmp(token, "INPUT") == 0) {
          mode = strdup("r");
        } else if (strcasecmp(token, "OUTPUT") == 0) {
          mode = strdup("w");
///kw+FILE+usage [ OPEN | DO_NOT_OPEN ]
        } else if (strcasecmp(token, "OPEN") == 0) {
          open = 1;
        } else if (strcasecmp(token, "DO_NOT_OPEN") == 0) {
          open = 0;
        } else {
          wasora_push_error_message("unknown keyword '%s'", token);
          return WASORA_PARSER_ERROR;
        }

      }

      if (mode == NULL) {
        wasora_push_error_message("file '%s' needs to have a mode or to be declared either as input or output", name);
        return WASORA_PARSER_ERROR;
      }

      if ((file = wasora_define_file(name, format, n_args, arg, mode, open)) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      if (wasora_define_instruction(wasora_instruction_file, file) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      free(format);
      free(mode);
      free(name);
      return WASORA_PARSER_OK;

// ---- CLOSE ----------------------------------------------------
///kw+CLOSE+usage CLOSE
    } else if (strcasecmp(token, "CLOSE") == 0) {
      
      file_t *file;
      
      wasora_call(wasora_parser_file(&file) != WASORA_PARSER_OK);
      
      if (wasora_define_instruction(wasora_instruction_close_file, file) == NULL) {
        return WASORA_PARSER_ERROR;
      }
      return WASORA_PARSER_OK;
      

// --- IF -----------------------------------------------------
///kw+IF+desc Begins a conditional block.
///kw+IF+usage IF expr
///kw+IF+usage &nbsp;
///kw+IF+usage  <block_of_instructions_if_expr_is_true>
///kw+IF+usage &nbsp;
///kw+IF+usage [ ELSE ]
///kw+IF+usage &nbsp;
///kw+IF+usage  [block_of_instructions_if_expr_is_false]
///kw+IF+usage &nbsp;
///kw+IF+usage ENDIF
    } else if (strcasecmp(token, "IF") == 0) {

      conditional_block_t *conditional_block;
      conditional_block = calloc(1, sizeof(conditional_block_t));
      LL_APPEND(wasora.conditional_blocks, conditional_block);

      conditional_block->condition = malloc(sizeof(expr_t));
      if (wasora_parser_expression(conditional_block->condition) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      if (wasora_get_next_token(NULL) != NULL) {
        wasora_push_error_message("conditional blocks should start in a separate line");
        return WASORA_PARSER_ERROR;
      }

      conditional_block->father = wasora.active_conditional_block;

      if (wasora_define_instruction(wasora_instruction_if, conditional_block) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      wasora.active_conditional_block = conditional_block;

      return WASORA_PARSER_OK;

    // ---- ELSE  ----------------------------------------------------
    } else if (strcasecmp(token, "ELSE") == 0) {

      instruction_t *instruction;

      conditional_block_t *conditional_block;
      conditional_block = calloc(1, sizeof(conditional_block_t));
      LL_APPEND(wasora.conditional_blocks, conditional_block);

      if (wasora.active_conditional_block->else_of != NULL) {
        wasora_push_error_message("more than one ELSE clause for a single IF clause", token);
        return WASORA_PARSER_ERROR;
      }

      if ((instruction = wasora_define_instruction(wasora_instruction_else, conditional_block)) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      wasora.active_conditional_block->first_false_instruction = instruction;

      conditional_block->else_of = wasora.active_conditional_block;
      conditional_block->father = wasora.active_conditional_block->father;
      wasora.active_conditional_block = conditional_block;

      return WASORA_PARSER_OK;

// ---- ENDIF  ----------------------------------------------------
    } else if (strcasecmp(token, "ENDIF") == 0) {

      instruction_t *instruction;

      if ((instruction = wasora_define_instruction(wasora_instruction_endif, wasora.active_conditional_block)) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      if (wasora.active_conditional_block->else_of == NULL) {
        wasora.active_conditional_block->first_false_instruction = instruction;
      } else {
        wasora.active_conditional_block->first_true_instruction = instruction;
      }

      wasora.active_conditional_block = wasora.active_conditional_block->father;

      return WASORA_PARSER_OK;

// ---- SEMAPHORE ----------------------------------------------------
///kw+SEMAPHORE+desc Performs either a wait or a post operation on a named shared semaphore.
///kw+SEMAPHORE+usage [ SEMAPHORE | SEM ]
    } else if ((strcasecmp(token, "SEMAPHORE") == 0) || (strcasecmp(token, "SEM") == 0)) {

      struct semaphore_t *semaphore;
      semaphore = calloc(1, sizeof(struct semaphore_t));
      LL_APPEND(wasora.semaphores, semaphore);

///kw+SEMAPHORE+usage <name>
      if (wasora_parser_string(&semaphore->name) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

///kw+SEMAPHORE+usage { WAIT | POST }
      char *keywords[] = {"WAIT", "POST", ""};
      int values[] = {semaphore_wait, semaphore_post, 0};
      wasora_call(wasora_parser_keywords_ints(keywords, values, (int *)&semaphore->operation));

      if (wasora_define_instruction(wasora_instruction_sem, semaphore) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// ---- READ / WRITE  ----------------------------------------------------
///kw+READ+desc  Reads data (variables, vectors o matrices) from external objects.
///kw+WRITE+desc Writes data (variables, vectors o matrices) to external objects.
///kw+WRITE+desc See the `READ` keyword for usage details.
///kw+READ+usage [ READ | WRITE ]
    } else if ((strcasecmp(token, "READ") == 0) || (strcasecmp(token, "WRITE") == 0)) {

      io_t *io;

      expr_t min;
      expr_t max;

      io = calloc(1, sizeof(io_t));
      LL_APPEND(wasora.ios, io);

      if (strcasecmp(token, "READ") == 0) {
        io->direction = io_read;
      } else if (strcasecmp(token, "WRITE") == 0) {
        io->direction = io_write;
      }

      while ((token = wasora_get_next_token(NULL)) != NULL) {
        // ---- SHM_OBJECT ---------------------------------------------------------
///kw+READ+usage [ SHM <name> ]
        if (strcasecmp(token, "SHM") == 0 || strcasecmp(token, "SHM_OBJECT") == 0) {

          io->type = io_shm;

          if (wasora_parser_string(&io->shm_name) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

        // ---- FILE_PATH ----------------------------------------------------
///kw+READ+usage [ { ASCII_FILE_PATH | BINARY_FILE_PATH } <path> ]
        } else if (strcasecmp(token, "ASCII_FILE_PATH") == 0 || strcasecmp(token, "BINARY_FILE_PATH") == 0) {

          if (strcasecmp(token, "ASCII_FILE_PATH") == 0) {
            io->type = io_file_ascii;
          } else if (strcasecmp(token, "BINARY_FILE_PATH") == 0) {
            io->type = io_file_binary;
          }

          if (wasora_parser_string(&io->filepath) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

        // ---- FILE ----------------------------------------------------
///kw+READ+usage [ { ASCII_FILE | BINARY_FILE } <identifier> ]
        } else if (strcasecmp(token, "ASCII_FILE") == 0 || strcasecmp(token, "BINARY_FILE") == 0) {

          if (strcasecmp(token, "ASCII_FILE") == 0) {
            io->type = io_file_ascii;
          } else if (strcasecmp(token, "BINARY_FILE") == 0) {
            io->type = io_file_binary;
          }

          if (wasora_parser_file(&io->file) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

          // ---- FILE ----------------------------------------------------
  ///kw+READ+usage [ IGNORE_NULL ]
          } else if (strcasecmp(token, "IGNORE_NULL") == 0) {

            io->ignorenull = 1;

          } else  {
///kw+READ+usage [ object_1 object_2 ... object_n ]

          // cosas
          io->n_things++;
          io_thing_t *thing = calloc(1, sizeof(io_thing_t));
          LL_APPEND(io->things, thing);

          // si pusieron un corchete quieren un pedazo solamente
          if ((dummy = strchr(token, '<')) != NULL) {
            *dummy = '\0';
          }

          if ((thing->variable = wasora_get_variable_ptr(token)) != NULL) {

            io->size += 1;

          } else if ((thing->vector = wasora_get_vector_ptr(token)) != NULL) {

            if (dummy == NULL) {
              thing->row_min = 0;
              thing->row_max = thing->vector->size;
            } else {
              *dummy = '<';
              if (wasora_parse_range(dummy, '<', ':', '>', &min, &max) != WASORA_PARSER_OK) {
                return WASORA_PARSER_ERROR;
              }
              thing->row_min = (int)wasora_evaluate_expression(&min)-1;
              thing->row_max = (int)wasora_evaluate_expression(&max);
              if (thing->row_max < thing->row_min) {
                wasora_push_error_message("mix row '%d' is greater that max row '%d", thing->row_min, thing->row_max);
                return WASORA_PARSER_ERROR;
              }
            }

            io->size += thing->row_max - thing->row_min;

          } else if ((thing->matrix = wasora_get_matrix_ptr(token)) != NULL) {

            if (dummy == NULL) {
              thing->row_min = 0;
              thing->row_max = thing->matrix->rows;
              thing->col_min = 0;
              thing->col_max = thing->matrix->cols;
            } else {
              *dummy = '<';
              if (wasora_parse_range(dummy, '<', ':', ';', &min, &max) != WASORA_PARSER_OK) {
                return WASORA_PARSER_ERROR;
              }
              thing->row_min = (int)wasora_evaluate_expression(&min)-1;
              thing->row_max = (int)wasora_evaluate_expression(&max);
              if (thing->row_max < thing->row_min) {
                wasora_push_error_message("mix row '%d' is greater that max row '%d", thing->row_min, thing->row_max);
                return WASORA_PARSER_ERROR;
              }

              if (wasora_parse_range(strchr(dummy, ';'), ';', ':', '>', &min, &max) != WASORA_PARSER_OK) {
                return WASORA_PARSER_ERROR;
              }

              thing->col_min = (int)wasora_evaluate_expression(&min)-1;
              thing->col_max = (int)wasora_evaluate_expression(&max);
              if (thing->col_max < thing->col_min) {
                wasora_push_error_message("mix col '%d' is greater that max col '%d", thing->col_min, thing->col_max);
                return WASORA_PARSER_ERROR;
              }

            }

            io->size += (thing->row_max-thing->row_min)*(thing->col_max-thing->col_min);

          } else {
            
            if (wasora_parse_expression(token, &thing->expr) != WASORA_PARSER_OK) {
              wasora_push_error_message("undefined keyword, variable, vector, matrix, alias or invalid expression '%s'", token);
              return WASORA_PARSER_ERROR;
            }
            
            if (io->direction == io_read) {
              wasora_push_error_message("expressions cannot be used in a READ instruction", token);
              return WASORA_PARSER_ERROR;
            }
            
            io->size += 1;
            
          }
        }
      }


      if (io->type == io_undefined)  {
        wasora_push_error_message("undefined I/O resource type");
        return WASORA_PARSER_ERROR;
      }

      if (io->type == io_shm && io->size == 0)  {
        wasora_push_error_message("empty shared memory object");
        return WASORA_PARSER_ERROR;
      }

      if (wasora_define_instruction(wasora_instruction_io, io) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// --- PRINT -----------------------------------------------------
///kw+PRINT+desc Prints plaint-text data to the standard output or to an output file.
///kw+PRINT+usage PRINT
    } else if (strcasecmp(token, "PRINT") == 0) {

      print_t *print;
      matrix_t *dummy_matrix;
      vector_t *dummy_vector;
      int n;

      print = calloc(1, sizeof(print_t));
      LL_APPEND(wasora.prints, print);

      char *keywords[] = {"SKIP_STEP", "SKIP_STATIC_STEP", "SKIP_TIME", "SKIP_HEADER_STEP", ""};
      expr_t *expressions[] = {
        &print->skip_step,
        &print->skip_static_step,
        &print->skip_time,
        &print->skip_header_step,
        NULL
      };

///kw+PRINT+detail Each argument `object` that is not a keyword is expected to be part of the output,
///kw+PRINT+detail can be either a matrix, a vector, an scalar algebraic expression. If the given object
///kw+PRINT+detail cannot be solved into a valid matrix, vector or expression, it is treated as a
///kw+PRINT+detail string literal if `IMPLICIT` is `ALLOWED`, otherwise a parser error is raised.
///kw+PRINT+detail To explicitly interpret an object as a literal string even if it resolves to a valid
///kw+PRINT+detail numerical expression, it should be prefixed with the `TEXT` keyword. Hashes~`#`
///kw+PRINT+detail appearing literal strings have to be quoted to prevent the parser to treat them as
///kw+PRINT+detail comments within the wasora input file.
///kw+PRINT+detail Whenever an argument starts with a porcentage sign '%', it is treated as a
///kw+PRINT+detail printf-compatible format definition and all the objects that follow it are printed
///kw+PRINT+detail using the given format until a new format definition is found. All the objects are
///kw+PRINT+detail treated as double-precision floating point numbers, so only floating point formats      
///kw+PRINT+detail should be given. The default format is '%e'.
///kw+PRINT+detail Matrices, vectors, scalar expressions, format modifiers and string literals can be
///kw+PRINT+detail given in the desired order, and are processed from left to right.
///kw+PRINT+detail Vectors are printed element-by-element in a single row. See `PRINT_VECTOR` to print
///kw+PRINT+detail vectors column-wise. Matrices are printed element-by-element in a row-major fashion
///kw+PRINT+detail if mixed with other objects but in the natural row and column fashion if it is the
///kw+PRINT+detail only given object.      
      
      while ((token = wasora_get_next_token(NULL)) != NULL) {
///kw+PRINT+usage [ FILE <identifier> ]
///kw+PRINT+detail If the `FILE` keyword is not provided, default is to write to stdout.
        if (strcasecmp(token, "FILE") == 0) {
          if (wasora_parser_file(&print->file) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+PRINT+usage [ NONEWLINE ]
        } else if (strcasecmp(token, "NONEWLINE") == 0) {
///kw+PRINT+detail If the `NONEWLINE` keyword is not provided, default is to write a newline '\\n' character after all the objects are processed.
          print->nonewline = 1;

///kw+PRINT+usage [ SEP <string> ]
        } else if (strcasecmp(token, "SEP") == 0 || strcasecmp(token, "SEPARATOR") == 0) {
          if (wasora_parser_string(&print->separator) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }
///kw+PRINT+detail The `SEP` keywords expects a string used to separate printed objects, the default is a tab '\\t' character.

///kw+PRINT+usage [ NOSEP ]
        } else if (strcasecmp(token, "NOSEP") == 0) {
          print->separator = strdup("");
///kw+PRINT+detail Use the `NOSEP` keyword to define an empty string as object separator.
          
///kw+PRINT+usage [ HEADER ]
        } else if (strcasecmp(token, "HEADER") == 0) {
          print->header = 1;
///kw+PRINT+detail If the `HEADER` keyword is given, a single line containing the literal text
///kw+PRINT+detail given for each object is printed at the very first time the `PRINT` instruction is
///kw+PRINT+detail processed, starting with a hash `#` character.           

///kw+PRINT+usage [ SKIP_STEP <expr> ]
///kw+PRINT+detail If the `SKIP_STEP` (`SKIP_STATIC_STEP`)keyword is given, the instruction is processed only every
///kw+PRINT+detail the number of transient (static) steps that results in evaluating the expression,
///kw+PRINT+detail which may not be constant. By default the `PRINT` instruction is evaluated every
///kw+PRINT+detail step. The `SKIP_HEADER_STEP` keyword works similarly for the optional `HEADER` but
///kw+PRINT+detail by default it is only printed once. The `SKIP_TIME` keyword use time advancements
///kw+PRINT+detail to choose how to skip printing and may be useful for non-constant time-step problems.
          
///kw+PRINT+usage [ SKIP_STATIC_STEP <expr> ]
///kw+PRINT+usage [ SKIP_TIME <expr> ]
///kw+PRINT+usage [ SKIP_HEADER_STEP <expr> ]
        } else if ((n = wasora_parser_match_keyword_expression(token, keywords, expressions, sizeof(expressions)/sizeof(expr_t *))) != WASORA_PARSER_UNHANDLED) {
          if (n == WASORA_PARSER_ERROR) {
            return WASORA_PARSER_ERROR;
          }

        } else {
///kw+PRINT+usage [ <object_1> <object_2> ... <object_n> ]
///kw+PRINT+usage [ TEXT <string_1> ... TEXT <string_n> ]
          // es un formato, una expresion o texto
          // asi que agregamos un eslabon a la lista de tokens
          print_token_t *print_token = calloc(1, sizeof(print_token_t));
          LL_APPEND(print->token, print_token);

          if (token[0] == '%') {
            print_token->format = strdup(token);

          } else if (strcasecmp(token, "STRING") == 0 || strcasecmp(token, "TEXT") == 0) {
            if (wasora_parser_string(&print_token->text) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }

          } else if ((dummy_matrix = wasora_get_matrix_ptr(token)) != NULL) {
            print_token->text = strdup(token);   // nos quedamos con el texto para el header
            print_token->matrix = dummy_matrix;

          } else if ((dummy_vector = wasora_get_vector_ptr(token)) != NULL) {
            print_token->text = strdup(token);   // nos quedamos con el texto para el header
            print_token->vector = dummy_vector;

          } else {
            if (wasora_parse_expression(token, &print_token->expression) != WASORA_PARSER_OK) {

              // dejamos que las expresiones no resueltas se transformen en texto
              // solo si no tenemos implicit none
              if (wasora.implicit_none) {
                wasora_push_error_message("implicit definition is not allowed");
                return WASORA_PARSER_ERROR;
              } else {
                print_token->expression.n_tokens = 0;
                print_token->text = strdup(token);
                wasora_pop_error_message();
              }
            } else {
              print_token->text = strdup(token);   // nos quedamos con el texto para el header
            }
          }
        }
      }

      // llenamos defaults
      if (print->separator == NULL) {
        print->separator = strdup(DEFAULT_PRINT_SEPARATOR);
      }

      if (print->file == NULL) {
        print->file = wasora.special_files.stdout_;
      }

      if (wasora_define_instruction(wasora_instruction_print, print) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;


// --- PRINT_FUNCTION -----------------------------------------------------
///kw+PRINT_FUNCTION+desc Prints one or more functions as a table of values of dependent and independent variables.
///kw+PRINT_FUNCTION+usage PRINT_FUNCTION
///kw+PRINT_FUNCTION+usage <function_1> [ { function_2 | expr_1 } ... { function_n | expr_n-1 } ]
    } else if (strcasecmp(token, "PRINT_FUNCTION") == 0) {

      print_function_t *print_function;
      function_t *dummy_function;

      print_function = calloc(1, sizeof(print_function_t));
      LL_APPEND(wasora.print_functions, print_function);


      while ((token = wasora_get_next_token(NULL)) != NULL) {
///kw+PRINT_FUNCTION+usage [ FILE <identifier> ]
        if (strcasecmp(token, "FILE") == 0) {
          if (wasora_parser_file(&print_function->file) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+PRINT_FUNCTION+usage [ HEADER ]
        } else if (strcasecmp(token, "HEADER") == 0) {
          print_function->header = 1;
          
///kw+PRINT_FUNCTION+usage [ MIN <expr_1> <expr_2> ... <expr_m> ]
        } else if (strcasecmp(token, "MIN") == 0) {

          if (print_function->first_function == NULL) {
            wasora_push_error_message("MIN before actual function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          if (wasora_parser_expressions(&print_function->range.min, print_function->first_function->n_arguments) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }


///kw+PRINT_FUNCTION+usage [ MAX <expr_1> <expr_2> ... <expr_m> ]
        } else if (strcasecmp(token, "MAX") == 0) {

          if (print_function->first_function == NULL) {
            wasora_push_error_message("MAX before actual function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          if (wasora_parser_expressions(&print_function->range.max, print_function->first_function->n_arguments) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+PRINT_FUNCTION+usage [ STEP <expr_1> <expr_2> ... <expr_m> ]
        } else if (strcasecmp(token, "STEP") == 0) {

          if (print_function->first_function == NULL) {
            wasora_push_error_message("STEP before actual function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          if (wasora_parser_expressions(&print_function->range.step, print_function->first_function->n_arguments) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+PRINT_FUNCTION+usage [ FORMAT <print_format> ]
        } else if (strcasecmp(token, "FORMAT") == 0) {
          if (wasora_parser_string(&print_function->format) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

        } else {

          // agregamos un eslabon a la lista de tokens
          print_token_t *print_token = calloc(1, sizeof(print_token_t));
          LL_APPEND(print_function->tokens, print_token);

          print_token->text = strdup(token);   // nos quedamos con el texto para el header
          
          if ((dummy_function = wasora_get_function_ptr(token)) != NULL) {
            print_token->function = dummy_function;
            if (print_function->first_function == NULL) {
              // es la primer funcion
              print_function->first_function = dummy_function;

            } else if (dummy_function->n_arguments !=  print_function->first_function->n_arguments) {
              wasora_push_error_message("functions do not have the same number of arguments");
              return WASORA_PARSER_ERROR;

            }

          } else {
            if (wasora_parse_expression(token, &print_token->expression) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }

          }
        }
      }

      if (print_function->first_function == NULL) {
        wasora_push_error_message("at least one function expected");
        return WASORA_PARSER_ERROR;
      }
      
      if (print_function->separator == NULL) {
        print_function->separator = strdup(DEFAULT_PRINT_SEPARATOR);
      }

      if (print_function->format == NULL) {
        print_function->format = strdup(DEFAULT_PRINT_FORMAT);
      }

      if (print_function->file == NULL) {
        print_function->file = wasora.special_files.stdout_;
      }

      if (wasora_define_instruction(wasora_instruction_print_function, print_function) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// --- PRINT_VECTOR -----------------------------------------------------
///kw+PRINT_VECTOR+desc Prints one or more vectors.
///kw+PRINT_VECTOR+usage PRINT_VECTOR
    } else if (strcasecmp(token, "PRINT_VECTOR") == 0) {

      print_vector_t *print_vector;

      print_vector = calloc(1, sizeof(print_vector_t));
      LL_APPEND(wasora.print_vectors, print_vector);

      while ((token = wasora_get_next_token(NULL)) != NULL) {

///kw+PRINT_VECTOR+usage [ FILE <identifier> ]
        if (strcasecmp(token, "FILE") == 0) {
          if (wasora_parser_file(&print_vector->file) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+PRINT_VECTOR+usage [ { VERTICAL | HORIZONTAL } ]
  } else if (strcasecmp(token, "VERTICAL") == 0) {
          print_vector->horizontal = 0;

        } else if (strcasecmp(token, "HORIZONTAL") == 0) {
          print_vector->horizontal = 1;

///kw+PRINT_VECTOR+usage [ ELEMS_PER_LINE <expr> ]
        } else if (strcasecmp(token, "ELEMS_PER_LINE") == 0) {
          print_vector->horizontal = 1;
          if (wasora_parser_expression(&print_vector->elems_per_line) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+PRINT_VECTOR+usage [ FORMAT <print_format> ]
        } else if (strcasecmp(token, "FORMAT") == 0) {
          if (wasora_parser_string(&print_vector->format) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

        } else {

///kw+PRINT_VECTOR+usage <vector_1> [ vector_2 ... vector_n ]
          // agregamos un eslabon a la lista de tokens
          print_token_t *print_token = calloc(1, sizeof(print_token_t));
          LL_APPEND(print_vector->tokens, print_token);

          if ((print_token->vector = wasora_get_vector_ptr(token)) != NULL) {
            if (print_vector->first_vector == NULL) {
              // es el primer vector
              print_vector->first_vector = print_token->vector;
            } else if (print_token->vector->size != print_vector->first_vector->size) {
              wasora_push_error_message("vectors do not have the same size");
              return WASORA_PARSER_ERROR;
            }

          } else {
            if (wasora_parse_expression(token, &print_token->expression) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }

          }
        }
      }


      // llenamos defaults
      if (print_vector->format == NULL) {
        print_vector->format = strdup(DEFAULT_PRINT_FORMAT);
      }

      if (print_vector->separator == NULL) {
        print_vector->separator = strdup(DEFAULT_PRINT_SEPARATOR);
      }

      if (print_vector->file == NULL) {
        print_vector->file = wasora.special_files.stdout_;
      }

      if (wasora_define_instruction(wasora_instruction_print_vector, print_vector) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// --- TEMPLATE -----------------------------------------------------
///kw+TEMPLATE+desc Replaces wildcards in a file with wasora expressions.
///kw+TEMPLATE+usage TEMPLATE TODO!!!
    } else if (strcasecmp(token, "TEMPLATE") == 0) {

      templ_t *templ;
      templ = calloc(1, sizeof(templ_t));
      LL_APPEND(wasora.templs, templ);

      if (wasora_parser_string(&templ->filename_input) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      if (wasora_parser_string(&templ->filename_output) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      i = 0;
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        if (strcasecmp(token, "WILDCARD") == 0) {

          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected wildcard character");
            return WASORA_PARSER_ERROR;
          }

          if (strlen(token) != 1) {
            wasora_push_error_message("wildcard has to be exactly one character");
            return WASORA_PARSER_ERROR;
          }

          templ->wildcard = token[0];

        } else {

          print_token_t *print_token = calloc(1, sizeof(print_token_t));
          LL_APPEND(templ->print_tokens, print_token);

          if (token[0] == '%') {
            print_token->format = strdup(token);

          } else {
            if (wasora_parse_expression(token, &print_token->expression) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }
            print_token->index = i++;

          }
        }

      }


      if (templ->print_tokens == NULL) {
        wasora_push_error_message("expected at least one expression");
        return WASORA_PARSER_ERROR;
      }

      if (templ->wildcard == '\0') {
        templ->wildcard = DEFAULT_TEMPLATE_WILDCARD;
      }

      if (wasora_define_instruction(wasora_instruction_templ,  templ) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;


// --- SHELL -----------------------------------------------------
///kw+SHELL+desc Executes a shell command.
///kw+SHELL+usage SHELL
    } else if (strcasecmp(token, "SHELL") == 0) {

      shell_t *shell;
      shell = calloc(1, sizeof(shell_t));
      LL_APPEND(wasora.shells, shell);

///kw+SHELL+usage <print_format> [ expr_1 expr_2 ... expr_n ]
      if (wasora_parser_string_format(&shell->format, &shell->n_args) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      shell->arg = calloc(shell->n_args, sizeof(expr_t));
      for (i = 0; i < shell->n_args; i++) {
        if (wasora_parser_expression(&shell->arg[i]) != WASORA_PARSER_OK) {
          return WASORA_PARSER_ERROR;
        }
      }

      if (wasora_define_instruction(wasora_instruction_shell, shell) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// --- CALL -----------------------------------------------------
///kw+CALL+desc Executes a previously dynamically-loaded user-provided routine.
///kw+CALL+usage CALL <name>
    } else if (strcasecmp(token, "CALL") == 0) {

      call_t *call;
      call = calloc(1, sizeof(call_t));
      LL_APPEND(wasora.calls, call);


      if (wasora_parser_string(&call->name) != WASORA_PARSER_OK) {
        return WASORA_PARSER_ERROR;
      }

      if ((call->function = wasora_get_routine_ptr(call->name)) == NULL) {
        wasora_push_error_message("undefined routine '%s'", call->name);
        return WASORA_PARSER_ERROR;
      }

///kw+CALL+usage [ expr_1 expr_2 ... expr_n ]
      call->n_arguments = 0;
      while ((token = wasora_get_next_token(NULL)) != NULL) {

      expr_t *argument = calloc(1, sizeof(expr_t));
      LL_APPEND(call->arguments, argument);

        if (wasora_parse_expression(token, argument) != WASORA_PARSER_OK) {
          return WASORA_PARSER_ERROR;
        }
        call->n_arguments++;
      }

      call->x = malloc(call->n_arguments*sizeof(double));

      if (wasora_define_instruction(wasora_instruction_call, call) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

// ----- HISTORY  -----------------------------------------------------------------
///kw+HISTORY+desc Record the time history of a variable as a function of time.
///kw+HISTORY+usage HISTORY
    } else if ((strcasecmp(token, "HISTORY") == 0)) {

      history_t *history;
      history = calloc(1, sizeof(history_t));
      LL_APPEND(wasora.histories, history);

///kw+HISTORY+usage <variable>
      // el nombre de la variable
      if (wasora_parser_variable(&history->variable)) {
        return WASORA_PARSER_ERROR;
      }

///kw+HISTORY+usage <function>
      // el nombre de la funcion
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected function name");
        return WASORA_PARSER_ERROR;
      }

      if ((history->function = wasora_define_function(token, 1)) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      if (wasora_define_instruction(wasora_instruction_history, history) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

    // ----- PARAMETRIC  -----------------------------------------------------------------
///kw+PARAMETRIC+desc Sistematically sweeps a zone of the parameter space.
    } else if ((strcasecmp(token, "PARAMETRIC") == 0)) {

      varlist_t *varlist = NULL;
      varlist_t *varitem;
      
      double xi;
      // suponemos que type = linear
      wasora.parametric.type = parametric_linear;
      wasora.parametric_mode = 1;

///kw+PARAMETRIC+usage PARAMETRIC
///kw+PARAMETRIC+usage <var_1> [ ... <var_n> ]
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        
        if (strcasecmp(token, "TYPE") == 0) {
///kw+PARAMETRIC+usage [ TYPE { linear  | logarithmic  | random  | gaussianrandom  | sobol  | niederreiter  | halton  | reversehalton } ]
          char *keywords[] = {"linear", "logarithmic", "random", "gaussianrandom", "sobol", "niederreiter", "halton", "reversehalton", ""};
          int values[] = {parametric_linear,
                          parametric_logarithmic,
                          parametric_random,
                          parametric_gaussianrandom,
                          parametric_sobol,
                          parametric_niederreiter,
                          parametric_halton,
                          parametric_reversehalton,
                          0};
          
          wasora_call(wasora_parser_keywords_ints(keywords, values, (int *)&wasora.parametric.type));
          
///kw+PARAMETRIC+usage [ MIN <numexpr_1> ... <numexpr_n> ]
        } else if (strcasecmp(token, "MIN") == 0) {

          if (wasora.parametric.dimensions == 0) {
            wasora_push_error_message("MIN before variables, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          wasora_call(wasora_parser_expressions(&wasora.parametric.range.min, wasora.parametric.dimensions));

///kw+PARAMETRIC+usage [ MAX <numexpr_1> ... <numexpr_n> ]
        } else if (strcasecmp(token, "MAX") == 0) {

          if (wasora.parametric.dimensions == 0) {
            wasora_push_error_message("MAX before variables, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          wasora_call(wasora_parser_expressions(&wasora.parametric.range.max, wasora.parametric.dimensions));
          
///kw+PARAMETRIC+usage [ STEP <numexpr_1> ... <numexpr_n> ]
        } else if (strcasecmp(token, "STEP") == 0) {

          if (wasora.parametric.dimensions == 0) {
            wasora_push_error_message("STEP before variables, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          wasora_call(wasora_parser_expressions(&wasora.parametric.range.step, wasora.parametric.dimensions));

///kw+PARAMETRIC+usage [ NSTEPS <numexpr_1> ... <numexpr_n> ]
        } else if (strcasecmp(token, "NSTEPS") == 0) {

          if (wasora.parametric.dimensions == 0) {
            wasora_push_error_message("NSTEPS before variables, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          wasora_call(wasora_parser_expressions(&wasora.parametric.range.nsteps, wasora.parametric.dimensions));

///kw+PARAMETRIC+usage [ OUTER_STEPS <numexpr> ]
        } else if (strcasecmp(token, "OUTER_STEPS") == 0) {

          wasora_call(wasora_parser_expression_in_string(&xi));
          if ((wasora.parametric.outer_steps = (int)xi) <= 0) {
            wasora_push_error_message("OUTER_STEPS has to be positive");
            return WASORA_PARSER_ERROR;
          }

///kw+PARAMETRIC+usage [ OFFSET <numexpr> ]
        } else if (strcasecmp(token, "OFFSET") == 0) {
          wasora_call(wasora_parser_expression_in_string(&xi));
          if ((wasora.parametric.offset = (int)(round(xi))) < 0) {
            wasora_push_error_message("expected a non-negative offset");
            return WASORA_PARSER_ERROR;
          }
                    
        } else {

          if (wasora.parametric.range.min != NULL ||
              wasora.parametric.range.max != NULL ||
              wasora.parametric.range.step != NULL  ||
              wasora.parametric.range.nsteps != NULL) {

            wasora_push_error_message("already given MIN, MAX, STEP or NSTEP, cannot add '%s' to the parametric list", token);
            return WASORA_PARSER_ERROR;
          }
          
          
          varitem = calloc(1, sizeof(varlist_t));
          if ((varitem->var = wasora_get_variable_ptr(token)) == NULL) {
            if ((varitem->var = wasora_define_variable(token)) == NULL) {
              return WASORA_PARSER_ERROR;
            }
          } 
          
          LL_APPEND(varlist, varitem);
          wasora.parametric.dimensions++;

        }
      }

      if (wasora.parametric.dimensions == 0) {
        wasora_push_error_message("no parametric variables given");
        return WASORA_PARSER_ERROR;
      }
      
      wasora.parametric.variable = calloc(wasora.parametric.dimensions, sizeof(var_t *));
      wasora.parametric.min = calloc(wasora.parametric.dimensions, sizeof(double));
      wasora.parametric.max = calloc(wasora.parametric.dimensions, sizeof(double));
      wasora.parametric.step = calloc(wasora.parametric.dimensions, sizeof(double));
          
      varitem = varlist;
      for (i = 0; i < wasora.parametric.dimensions; i++) {
        wasora.parametric.variable[i] = varitem->var;
        if ((varitem = varitem->next) == NULL && i != wasora.parametric.dimensions-1) {
          wasora_push_error_message("internal mismatch in number of fit parameter %d", i);
          return WASORA_PARSER_ERROR;
        }
        
        if (wasora.parametric.range.min != NULL) {
          wasora.parametric.min[i] = wasora_evaluate_expression(&wasora.parametric.range.min[i]);
        } else {
          wasora.parametric.min[i] = 0.0;
        }
          

        if (wasora.parametric.range.max != NULL) {
          wasora.parametric.max[i] = wasora_evaluate_expression(&wasora.parametric.range.max[i]);
        } else {
          wasora.parametric.max[i] = 1.0;
        }
        
        // el step no importa
        if (wasora.parametric.range.step != NULL) {
          wasora.parametric.step[i] = wasora_evaluate_expression(&wasora.parametric.range.step[i]);
        } else if (wasora.parametric.range.nsteps != NULL) {
          wasora.parametric.step[i] = (wasora.parametric.max[i]-wasora.parametric.min[i])/(wasora_evaluate_expression(&wasora.parametric.range.nsteps[i]));
        }
        
        if (wasora.parametric.max[i] < wasora.parametric.min[i]) {
          wasora_push_error_message("min (%e) greater than max (%e)", wasora.parametric.min[i], wasora.parametric.max[i]);
          return WASORA_PARSER_ERROR;
        }
        if (wasora.parametric.max[i]-wasora.parametric.min[i] < wasora_value(wasora_special_var(zero))) {
          wasora_push_error_message("min and max too close");
          return WASORA_PARSER_ERROR;
        }
      }

      if (wasora.parametric.outer_steps == 0 && wasora.parametric.step[0] == 0) {
        wasora_push_error_message("neither STEP nor TOTAL_STEPS keywords given");
        return WASORA_PARSER_ERROR;
      }
      
      return WASORA_PARSER_OK;

    // ----- FIT  -----------------------------------------------------------------
    } else if ((strcasecmp(token, "FIT") == 0)) {

///kw+FIT+usage FIT
      char *sigma_name;
      varlist_t *varlist = NULL;
      varlist_t *varitem;

      wasora.fit_mode = 1;

///kw+FIT+usage <function_to_be_fitted> 
      // la funcion cuyos parametros vamos a ajustar
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected function name");
        return WASORA_PARSER_ERROR;
      }

      if ((wasora.fit.function = wasora_get_function_ptr(token)) == NULL) {
        if (strchr(token, '(') != NULL) {
          wasora_push_error_message("function '%s' undefined (only the function name is to be provided, do not include its arguments)", token);
        } else {
          wasora_push_error_message("function '%s' undefined", token);
        }
        return WASORA_PARSER_ERROR;
      }

///kw+FIT+usage TO <function_with_data>
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected function name");
        return WASORA_PARSER_ERROR;
      }
      if (strcasecmp(token, "TO") != 0) {
        wasora_push_error_message("expected keyword 'TO' instead of '%s'", token);
      }

// la funcion con los datos experimentales
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected function name");
        return WASORA_PARSER_ERROR;
      }
      if ((wasora.fit.data = wasora_get_function_ptr(token)) == NULL) {
        if (strchr(token, '(') != NULL) {
          wasora_push_error_message("function '%s' undefined (only the function name is to be provided, do not include its arguments)", token);
        } else {
          wasora_push_error_message("function '%s' undefined", token);
        }
        return WASORA_PARSER_ERROR;
      }

      if (wasora.fit.function->n_arguments != wasora.fit.data->n_arguments) {
        wasora_push_error_message("function '%s' has %d arguments and '%s' has %d", wasora.fit.function->name, wasora.fit.function->n_arguments, wasora.fit.data->name, wasora.fit.data->n_arguments);
        return WASORA_PARSER_ERROR;
      }

      if ((wasora.fit.n = wasora.fit.data->data_size) == 0) {
        wasora_push_error_message("function '%s' has to be point-wise defined", wasora.fit.data->name);
        return WASORA_PARSER_ERROR;
      }

///kw+FIT+usage VIA
      if ((token = wasora_get_next_token(NULL)) == NULL) {
        wasora_push_error_message("expected function name");
        return WASORA_PARSER_ERROR;
      }
      if (strcasecmp(token, "VIA") != 0) {
        wasora_push_error_message("expected keyword 'VIA' instead of '%s'", token);
        return WASORA_PARSER_ERROR;
      }

///kw+FIT+usage <var_1> <var_2> ... <var_n>
      while ((token = wasora_get_next_token(NULL)) != NULL) {

        // TODO: convergencia por gradiente, incertezas, covarianza, incrementos para gradiente automatico
///kw+FIT+usage [ GRADIENT <expr_1> <expr_2> ... <expr_n> ]
        if (strcasecmp(token, "GRADIENT") == 0) {
 
          if (wasora.fit.p == 0) {
            wasora_push_error_message("GRADIENT keyword before parameters");
            return WASORA_PARSER_ERROR;
          }
          
          wasora.fit.gradient = malloc(wasora.fit.p * sizeof(expr_t));
          for (i = 0; i < wasora.fit.p; i++) {
            if ((token = wasora_get_next_token(NULL)) == NULL) {
              wasora_push_error_message("expected an expression");
              return WASORA_PARSER_ERROR;
            }

            if (wasora_parse_expression(token, &wasora.fit.gradient[i]) != 0) {
              return WASORA_PARSER_ERROR;
            }
          }

///kw+FIT+usage [ RANGE_MIN <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "RANGE_MIN") == 0) {

          if (wasora.fit.data->n_arguments == 0) {
            wasora_push_error_message("RANGE_MIN before target function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          wasora_call(wasora_parser_expressions(&wasora.fit.range.min, wasora.fit.data->n_arguments));

///kw+FIT+usage [ RANGE_MAX <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "RANGE_MAX") == 0) {

          if (wasora.fit.data->n_arguments == 0) {
            wasora_push_error_message("RANGE_MAX before target function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          wasora_call(wasora_parser_expressions(&wasora.fit.range.max, wasora.fit.data->n_arguments));
          
///kw+FIT+usage [ DELTAEPSREL <expr> ]
        } else if (strcasecmp(token, "DELTAEPSREL") == 0) {

          wasora_call(wasora_parser_expression(&wasora.fit.deltaepsrel));

///kw+FIT+usage [ DELTAEPSABS <expr> ]
        } else if (strcasecmp(token, "DELTAEPSABS") == 0) {

          wasora_call(wasora_parser_expression(&wasora.fit.deltaepsabs));

///kw+FIT+usage [ MAX_ITER <expr> ]
        } else if (strcasecmp(token, "MAX_ITER") == 0) {

          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected an expression");
            return WASORA_PARSER_ERROR;
          }

          wasora.fit.max_iter = (int)wasora_evaluate_expression_in_string(token);
          if (wasora.fit.max_iter <= 0) {
            wasora_push_error_message("expected a positive integer");
            return WASORA_PARSER_ERROR;
          }

///kw+FIT+usage [ VERBOSE ]
        } else if (strcasecmp(token, "VERBOSE") == 0) {

          wasora.fit.verbose = 1;

///kw+FIT+usage [ RERUN | DO_NOT_RERUN ]
        } else if (strcasecmp(token, "DO_NOT_RERUN") == 0 || strcasecmp(token, "NORERUN") == 0) {

          wasora.fit.norerun = 1;

        } else if (strcasecmp(token, "RERUN") == 0) {

          wasora.fit.norerun = 0;

        } else {

          varitem = calloc(1, sizeof(varlist_t));
          if ((varitem->var = wasora_get_variable_ptr(token)) == NULL) {
            wasora_push_error_message("unknown variable '%s'", token);
            return WASORA_PARSER_ERROR;
          } 
          
          LL_APPEND(varlist, varitem);
          wasora.fit.p++;

        }
      }

      if (wasora.fit.p == 0) {
        wasora_push_error_message("no fit parameters given");
        return WASORA_PARSER_ERROR;
      }

      wasora.fit.param = malloc(wasora.fit.p * sizeof(var_t *));
      wasora.fit.sigma = malloc(wasora.fit.p * sizeof(var_t *));
      varitem = varlist;
      for (i = 0; i < wasora.fit.p; i++) {
        wasora.fit.param[i] = varitem->var;
        
        sigma_name = malloc(strlen(varitem->var->name)+strlen("sigma_")+8);
        sprintf(sigma_name, "sigma_%s", varitem->var->name);
        if ((wasora.fit.sigma[i] = wasora_define_variable(sigma_name)) == NULL) {
          return WASORA_PARSER_ERROR;
        }
        free(sigma_name);
        
        if ((varitem = varitem->next) == NULL && i != wasora.fit.p-1) {
          wasora_push_error_message("internal mismatch in number of fit parameter %d", i);
          return WASORA_PARSER_ERROR;
        }
      }

      if (wasora.fit.algorithm == NULL) {
        wasora.fit.algorithm = DEFAULT_NLIN_FIT_METHOD;
      }
      if (wasora.fit.max_iter == 0) {
        wasora.fit.max_iter = DEFAULT_NLIN_FIT_MAX_ITER;
      }

      return WASORA_PARSER_OK;

    // ----- OPTIMIZE  -----------------------------------------------------------------
    } else if ((strcasecmp(token, "OPTIMIZE") == 0) || (strcasecmp(token, "MINIMIZE") == 0)) {

      // ponemos esto en true porque normalmente no queremos que imprima todo el chorizo
      wasora.min_mode = 1;

///kw+OPTIMIZE+usage [ OPTIMIZE | MINIMIZE ]
      while ((token = wasora_get_next_token(NULL)) != NULL) {

///kw+OPTIMIZE+usage { FUNCTION <function> |
        if (strcasecmp(token, "FUNCTION") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.function = wasora_get_function_ptr(token)) == NULL) {
            if (strchr(token, '(') != NULL) {
              wasora_push_error_message("function '%s' undefined (only the function name is to be provided, do not include its arguments)", token);
            } else {
              wasora_push_error_message("function '%s' undefined", token);
            }
            return WASORA_PARSER_ERROR;
          }

          wasora.min.n = wasora.min.function->n_arguments;
          wasora.min.x = calloc(wasora.min.n, sizeof(var_t *));
          for (i = 0; i < wasora.min.n; i++) {
            wasora.min.x[i] = wasora.min.function->var_argument[i];
          }

///kw+OPTIMIZE+usage SIMAN_EFUNC <routine> }
        } else if (strcasecmp(token, "SIMAN_EFUNC") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_Efunc = (double (*)(void *)) wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ { METHOD | ALGORITHM } {
        } else if (strcasecmp(token, "METHOD") == 0 || strcasecmp(token, "ALGORITHM") == 0) {

          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected algorithm name");
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage conjugate_fr |
          if (strcasecmp(token, "conjugate_fr") == 0) {
            wasora.min.fdf_type = gsl_multimin_fdfminimizer_conjugate_fr;
///kw+OPTIMIZE+usage conjugate_pr |
          } else if (strcasecmp(token, "conjugate_pr") == 0) {
            wasora.min.fdf_type = gsl_multimin_fdfminimizer_conjugate_pr;
///kw+OPTIMIZE+usage vector_bfgs2 |
          } else if (strcasecmp(token, "vector_bfgs2") == 0) {
            wasora.min.fdf_type = gsl_multimin_fdfminimizer_vector_bfgs2;
///kw+OPTIMIZE+usage vector_bfgs |
          } else if (strcasecmp(token, "vector_bfgs") == 0) {
            wasora.min.fdf_type = gsl_multimin_fdfminimizer_vector_bfgs;
///kw+OPTIMIZE+usage steepest_descent |
          } else if (strcasecmp(token, "steepest_descent") == 0) {
            wasora.min.fdf_type = gsl_multimin_fdfminimizer_steepest_descent;
///kw+OPTIMIZE+usage nmsimplex2 |
          } else if (strcasecmp(token, "nmsimplex2") == 0) {
            wasora.min.f_type = gsl_multimin_fminimizer_nmsimplex2;
///kw+OPTIMIZE+usage nmsimplex |
          } else if (strcasecmp(token, "nmsimplex") == 0) {
            wasora.min.f_type = gsl_multimin_fminimizer_nmsimplex;
///kw+OPTIMIZE+usage nmsimplex2rand |
          } else if (strcasecmp(token, "nmsimplex2rand") == 0) {
            wasora.min.f_type = gsl_multimin_fminimizer_nmsimplex2rand;
///kw+OPTIMIZE+usage siman }
          } else if (strcasecmp(token, "siman") == 0) {
            wasora.min.siman_type = gsl_rng_default;
/*
          } else if (strcasecmp(token, "genetic") == 0) {
            wasora.min.genetic = 1;
          } else if (strcasecmp(token, "pure_genetic") == 0) {
            wasora.min.genetic = 2;
*/
          } else {
            wasora_push_error_message("unknown minimization method '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ GRADIENT <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "GRADIENT") == 0) {

          wasora.min.gradient = malloc(wasora.min.n * sizeof(expr_t));
          for (i = 0; i < wasora.min.n; i++) {
            if ((token = wasora_get_next_token(NULL)) == NULL) {
              wasora_push_error_message("expected an expression");
              return WASORA_PARSER_ERROR;
            }

            if (wasora_parse_expression(token, &wasora.min.gradient[i]) != 0) {
              return WASORA_PARSER_ERROR;
            }
          }

///kw+OPTIMIZE+usage [ GUESS <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "GUESS") == 0) {

          wasora.min.guess = malloc(wasora.min.n * sizeof(expr_t));
          for (i = 0; i < wasora.min.n; i++) {
            if ((token = wasora_get_next_token(NULL)) == NULL) {
              wasora_push_error_message("expected an expression");
              return WASORA_PARSER_ERROR;
            }

            if (wasora_parse_expression(token, &wasora.min.guess[i]) != 0) {
              return WASORA_PARSER_ERROR;
            }
          }

///kw+OPTIMIZE+usage [ MIN <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "MIN") == 0) {

          if (wasora.min.n == 0) {
            wasora_push_error_message("MIN before target function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          if (wasora_parser_expressions(&wasora.min.range.min, wasora.min.n) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ MAX <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "MAX") == 0) {

          if (wasora.min.n == 0) {
            wasora_push_error_message("MAX before target function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }

          if (wasora_parser_expressions(&wasora.min.range.max, wasora.min.n) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ STEP <expr_1> <expr_2> ... <expr_n> ]
        } else if (strcasecmp(token, "STEP") == 0) {

          if (wasora.min.n == 0) {
            wasora_push_error_message("STEP before target function, cannot determine number of arguments");
            return WASORA_PARSER_ERROR;
          }
          
          // los algoritmos con derivadas y siman necesitan un solo step
          // los sin derivadas necesitan n steps
          // si fdf_type no es null, entonces estamos con derivadas
          // igual despues chequeamos que no se mezclen
          wasora.min.n_steps = (wasora.min.f_type != NULL) ? wasora.min.n : 1;
          wasora.min.range.step = malloc(wasora.min.n_steps * sizeof(expr_t));

          if (wasora_parser_expressions(&wasora.min.range.step, wasora.min.n_steps) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_INIT <routine> ]
        } else if (strcasecmp(token, "SIMAN_INIT") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_init = (void *(*) (void))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_STEP <routine> ]
        } else if (strcasecmp(token, "SIMAN_STEP") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
            }
          if ((wasora.min.siman_step = (void (*) (const gsl_rng *, void *, double))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_COPY <routine> ]
        } else if (strcasecmp(token, "SIMAN_COPY") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_copy = (void (*) (void *, void *))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_COPY_CONSTRUCT <routine> ]
        } else if (strcasecmp(token, "SIMAN_COPY_CONSTRUCT") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_copy_construct = (void *(*) (void *))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_PRINT <routine> ]
        } else if (strcasecmp(token, "SIMAN_PRINT") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_print = (void (*) (void *))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_BEST <routine> ]
        } else if (strcasecmp(token, "SIMAN_BEST") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_best = (void (*) (void *))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ SIMAN_DESTROY <routine> ]
        } else if (strcasecmp(token, "SIMAN_DESTROY") == 0) {
          if ((token = wasora_get_next_token(NULL)) == NULL) {
            wasora_push_error_message("expected a routine");
            return WASORA_PARSER_ERROR;
          }
          if ((wasora.min.siman_destroy = (void (*) (void *))wasora_get_routine_ptr(token)) == NULL) {
            wasora_push_error_message("unknown routine '%s'", token);
            return WASORA_PARSER_ERROR;
          }

///kw+OPTIMIZE+usage [ VERBOSE ]
        } else if (strcasecmp(token, "VERBOSE") == 0) {

          wasora.min.verbose = 1;

///kw+OPTIMIZE+usage [ NORERUN ]
        } else if (strcasecmp(token, "NORERUN") == 0) {

          wasora.min.norerun = 1;

        } else {

///kw+OPTIMIZE+usage [ MAX_ITER <expr> ]
///kw+OPTIMIZE+usage [ TOL <expr> ]
///kw+OPTIMIZE+usage [ GRADTOL <expr> ]
///kw+OPTIMIZE+usage [ N_TRIES <expr> ]
///kw+OPTIMIZE+usage [ K <expr> ]
///kw+OPTIMIZE+usage [ T_INIT <expr> ]
///kw+OPTIMIZE+usage [ T_MIN <expr> ]
///kw+OPTIMIZE+usage [ MU <expr> ]
          
          // en una sola linea para el generador del lexer de pygments
          char *keywords[] = {"MAX_ITER", "TOL", "GRADTOL", "TRIES", "K", "T_INIT", "T_MIN", "MU", ""};
          //, "POPULATION", "GA_STEPS"};
          expr_t *expressions[] = {
            &wasora.min.max_iter,
            &wasora.min.tol,
            &wasora.min.gradtol,
            &wasora.min.n_tries,
            &wasora.min.k,
            &wasora.min.t_init,
            &wasora.min.t_min,
            &wasora.min.mu,
//            &wasora.min.population,
//            &wasora.min.ga_steps
            NULL,
          };

          if (wasora_parser_match_keyword_expression(token, keywords, expressions, sizeof(expressions)/sizeof(expr_t *)) == WASORA_PARSER_UNHANDLED) {
            wasora_push_error_message("unknown keyword '%s'", token);
            return WASORA_PARSER_ERROR;
          }
        }

      }

      if (wasora.min.function == NULL && wasora.min.siman_Efunc == NULL) {
        wasora_push_error_message("nothing to optimize!");
        return WASORA_PARSER_ERROR;
      }

      if (wasora.min.genetic != 0) {
        if (wasora.min.range.min == NULL) {
          wasora_push_error_message("need a MIN keyword for genetic algorithms");
          return WASORA_PARSER_ERROR;
        }

        if (wasora.min.range.max == NULL) {
          wasora_push_error_message("need a MAX keyword for genetic algorithms");
          return WASORA_PARSER_ERROR;
        }
      }

      if (wasora.min.fdf_type == NULL && wasora.min.f_type == NULL
       && wasora.min.siman_type == NULL && wasora.min.genetic == 0) {
        wasora.min.f_type = DEFAULT_MINIMIZER_METHOD;
      }
/*
      if (wasora.min.max_iter.n_tokens == 0) {
        parse_expr(DEFAULT_MINIMIZER_MAX_ITER, &wasora.min.max_iter);
      }

      if (wasora.min.n_steps == 0) {
        wasora.min.n_steps = (wasora.min.fdf_type != NULL) ? 1 : wasora.min.n;
      } else {
        if (wasora.min.fdf_type != NULL && wasora.min.n_steps != 1) {
          wasora_push_error_message(only one step size is expected when using a derivative-based algorithm");
          return WASORA_PARSER_ERROR;
        }
        if (wasora.min.f_type != NULL && wasora.min.n_steps != wasora.min.n) {
          wasora_push_error_message(expected %d step sizes when using a non-derivative-based algorithm", wasora.min.n);
          return WASORA_PARSER_ERROR;
        }
      }
  */

      if (wasora.min.siman_Efunc != NULL) {
        if (wasora.min.siman_init == NULL) {
          wasora_push_error_message("a siman init routine has to be provided");
          return WASORA_PARSER_ERROR;
        }
        if (wasora.min.siman_step == NULL) {
          wasora_push_error_message("a siman step routine has to be provided");
          return WASORA_PARSER_ERROR;
        }
        if (wasora.min.siman_copy == NULL) {
          wasora_push_error_message("a siman copy routine has to be provided");
          return WASORA_PARSER_ERROR;
        }
        if (wasora.min.siman_copy_construct == NULL) {
          wasora_push_error_message("a siman copy_construct routine has to be provided");
          return WASORA_PARSER_ERROR;
        }
        if (wasora.min.siman_destroy == NULL) {
          wasora_push_error_message("a siman destroy routine has to be provided");
          return WASORA_PARSER_ERROR;
        }


      }

      return WASORA_PARSER_OK;

    // ----- PHASE_SPACE -----------------------------------------------------------
    } else if (strcasecmp(token, "PHASE_SPACE") == 0) {
///kw+PHASE_SPACE+usage PHASE_SPACE { <vars> | <vectors> | <matrices> }

      var_t *variable, *variable_dot;
      vector_t *vector, *vector_dot;
      matrix_t *matrix, *matrix_dot;
      char *buffer;
      int k, l;

      if (wasora_dae.dimension != 0) {
        wasora_push_error_message("PHASE_SPACE keyword already given");
        return WASORA_PARSER_ERROR;
      }
#ifdef HAVE_IDA
      if (sizeof(realtype) != sizeof(double)) {
        wasora_push_error_message("\nSUNDIALS was compiled using a different word size than wasora.\nPlease recompile with double precision floating point arithmetic.");
        return WASORA_PARSER_ERROR;
      }
#endif

      i = 0;
      while ((token = wasora_get_next_token(NULL)) != NULL) {

        phase_object_t *phase_object = calloc(1, sizeof(phase_object_t));
        LL_APPEND(wasora_dae.phase_objects, phase_object);

        if ((vector = wasora_get_vector_ptr(token)) != NULL) {

          phase_object->offset = i;
          phase_object->vector = vector;
          phase_object->size = vector->size;
          phase_object->name = vector->name;

          buffer = malloc(strlen(vector->name)+8);
          sprintf(buffer, "%s_dot", vector->name);
          vector_dot = wasora_define_vector(buffer, vector->size, NULL);
          free(buffer);

          wasora_dae.dimension += vector->size;
          wasora_dae.phase_value = realloc(wasora_dae.phase_value, (wasora_dae.dimension+1)*sizeof(double));
          wasora_dae.phase_derivative = realloc(wasora_dae.phase_derivative, (wasora_dae.dimension+1)*sizeof(double));

          for (k = 0; k < vector->size; k++) {
            wasora_dae.phase_value[i] = gsl_vector_ptr(wasora_value_ptr(vector), k);
            wasora_dae.phase_derivative[i] = gsl_vector_ptr(wasora_value_ptr(vector_dot), k);
            i++;
          }

        } else if ((matrix = wasora_get_matrix_ptr(token)) != NULL) {

          phase_object->offset = i;
          phase_object->matrix = matrix;
          phase_object->size = matrix->rows * matrix->cols;
          phase_object->name = matrix->name;

          buffer = malloc(strlen(matrix->name)+8);
          sprintf(buffer, "%s_dot", matrix->name);
          matrix_dot = wasora_define_matrix(buffer, matrix->rows, matrix->cols, NULL);
          free(buffer);

          wasora_dae.dimension += matrix->rows * matrix->cols;
          wasora_dae.phase_value = realloc(wasora_dae.phase_value, (wasora_dae.dimension+1)*sizeof(double));
          wasora_dae.phase_derivative = realloc(wasora_dae.phase_derivative, (wasora_dae.dimension+1)*sizeof(double));

          for (k = 0; k < matrix->rows; k++) {
            for (l = 0; l < matrix->cols; l++) {
              wasora_dae.phase_value[i] = gsl_matrix_ptr(wasora_value_ptr(matrix), k, l);
              wasora_dae.phase_derivative[i] = gsl_matrix_ptr(wasora_value_ptr(matrix_dot), k, l);
              i++;
            }
          }

        } else {

          if ((variable = wasora_get_or_define_variable_ptr(token)) == NULL) {
            return WASORA_PARSER_ERROR;
          }

          phase_object->offset = i;
          phase_object->var = variable;
          phase_object->size = 1;
          phase_object->name = variable->name;

          buffer = malloc(strlen(variable->name)+8);
          sprintf(buffer, "%s_dot", variable->name);
          variable_dot = wasora_define_variable(buffer);
          free(buffer);

          wasora_dae.dimension++;
          wasora_dae.phase_value = realloc(wasora_dae.phase_value, (wasora_dae.dimension+1)*sizeof(double));
          wasora_dae.phase_derivative = realloc(wasora_dae.phase_derivative, (wasora_dae.dimension+1)*sizeof(double));

          wasora_dae.phase_value[i] = wasora_value_ptr(variable);
          wasora_dae.phase_derivative[i] = wasora_value_ptr(variable_dot);
          i++;

        }
      }

      if (wasora_dae.dimension == 0) {
        wasora_push_error_message("empty phase space in PHASE_SPACE keyword");
        return WASORA_PARSER_ERROR;
      }

      // allocamos los residuos pero no los llenamos hasta las ecucaciones
      wasora_dae.residual = calloc(wasora_dae.dimension, sizeof(assignment_t));
      wasora_dae.equation_type = calloc(wasora_dae.dimension, sizeof(int));

      return WASORA_PARSER_OK;

    // ----- DIFFERENTIAL -----------------------------------------------------------
    } else if (strcasecmp(token, "DIFFERENTIAL") == 0) {
///kw+DIFFERENTIAL+usage DIFFERENTIAL { <vars> | <vectors> | <matrices> }
      
      phase_object_t *phase_object;
      
      while ((token = wasora_get_next_token(NULL)) != NULL) {
        LL_FOREACH(wasora_dae.phase_objects, phase_object) {
          if ((phase_object->var    != NULL && strcmp(token, phase_object->var->name) == 0) ||
              (phase_object->vector != NULL && strcmp(token, phase_object->vector->name) == 0) ||
              (phase_object->matrix != NULL && strcmp(token, phase_object->matrix->name) == 0) ) {
            phase_object->differential = 1;
          }
        }          
      }
      
      return WASORA_PARSER_OK;
      
      // --- DAE -----------------------------------------------------
    } else if (token[0] == '0' || strstr(wasora.line, ".=") != NULL) {

///kw+.=+usage { 0[(i[,j]][<imin:imax[;jmin:jmax]>] | <expr1> } .= <expr2>
      char *equation;
      char *dummy;
      char buffer[MAX_STRING];
      expr_t min, max;
      int i_min, i_max, j_min, j_max; 
      phase_object_t *phase_object;
      vector_t *vector = NULL;
      matrix_t *matrix = NULL;
      int j;

      if (wasora_dae.dimension == 0) {
        wasora_push_error_message("DAE equation given before defining the phase space");
        return WASORA_PARSER_ERROR;
      }

      if (wasora.i_dae >= wasora_dae.dimension) {
        wasora_push_error_message("more DAE equations than phase space dimension %d", wasora_dae.dimension);
        return WASORA_PARSER_ERROR;
      }

      if (token[0] == '0') {
        // ecuacion escrita en forma implicita
        wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_IMPLICIT;
        if ((dummy = strchr((char *)(token+strlen(token)+1), '=')) == NULL) {
          wasora_push_error_message("expecting equal sign after keyword 0");
          return WASORA_PARSER_ERROR;
        }
        equation = strdup(dummy+1);

        if (strncmp(token, "0(i,j)", 6) == 0) {
          //  ecuacion implicita y matricial
          wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_MATRICIAL;
          if ((matrix = wasora_get_first_matrix(equation)) == NULL) {
            wasora_push_error_message("matrix equations need at least one matrix in the expression");
            return WASORA_PARSER_ERROR;
          }
          
          if (token[6] == '<') {
            if (wasora_parse_range(token+6, '<', ':', ';', &min, &max) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }
            i_min = (int)wasora_evaluate_expression(&min)-1;
            i_max = (int)wasora_evaluate_expression(&max);
          
            if (wasora_parse_range(strchr(dummy+6, ';'), ';', ':', '>', &min, &max) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }
            j_min = (int)wasora_evaluate_expression(&min)-1;
            j_max = (int)wasora_evaluate_expression(&max);
          } else {
            i_min = 0;
            i_max = matrix->rows;
            j_min = 0;
            j_max = matrix->cols;
          }
          
        } else if (strncmp(token, "0(i)", 4) == 0) {
          // ecuacion implicita y vectorial
          wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_VECTORIAL;
          if ((vector = wasora_get_first_vector(equation)) == NULL) {
            wasora_push_error_message("vectorial equations need at least one vector in the expression");
            return WASORA_PARSER_ERROR;
          }

          if (token[4] == '<') {
            if (wasora_parse_range(token+4, '<', ':', '>', &min, &max) != WASORA_PARSER_OK) {
              return WASORA_PARSER_ERROR;
            }
            i_min = (int)wasora_evaluate_expression(&min)-1;
            i_max = (int)wasora_evaluate_expression(&max);
          } else {
            i_min = 0;
            i_max = vector->size;
          }
          
        } else {
          wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_SCALAR;
        }
        
        

      } else if (strstr(wasora.line, ".=") != NULL) {

        // el tipo puso una ecuacion explicita
        wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_EXPLICIT;

        if (strchr((char *)(token+strlen(token)+1), '=') == NULL) {
          wasora_push_error_message("syntax error, the equal sign should come right after the dot with no spaces between them");
          return WASORA_PARSER_ERROR;
        }
        dummy = strchr((char *)(token+strlen(token)+1), '=')+1;
        equation = malloc(strlen(dummy)+strlen(token)+16);

        // pasamos el miembro derecho para el otro lado cambiado de signo
        sprintf(equation, "(%s)-(%s)", token, dummy);

        // si en el miembro izquierdo hay un vector o una matriz, entonces araca
        if ((vector = wasora_get_first_vector(token)) != NULL) {
          wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_VECTORIAL;
          i_min = 0;
          i_max = vector->size;
        } else if ((matrix = wasora_get_first_matrix(token)) != NULL) {
          wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_MATRICIAL;
          i_min = 0;
          i_max = matrix->rows;
          j_min = 0;
          j_max = matrix->cols;
        } else {
          wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_SCALAR;
        }

      } else {
        return WASORA_PARSER_ERROR;
      }


      // vemos si es diferencial o algebraica: si aparece alguna derivada en la
      // expresion implicita, es diferencial
      if ((dummy = wasora_get_first_dot(equation)) != NULL) {
        int found = 0;
        wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_DIFFERENTIAL;
        LL_FOREACH(wasora_dae.phase_objects, phase_object) {
          if (strcmp(dummy, phase_object->name) == 0) {
            phase_object->differential = 1;
            found = 1;
          }
        }
        free(dummy);
        if (found == 0) {
          wasora_push_error_message("requested derivative of object '%s' but it is not in the phase space", dummy);
          return WASORA_PARSER_ERROR;
        }

      } else {
        wasora_dae.equation_type[wasora.i_dae] |= EQN_FLAG_ALGEBRAIC;
      }
      
      if (wasora.i_dae == 0) {
        // si es la primera dae, anotamos a que instruccion corresponde
        if ((wasora_dae.instruction = wasora_define_instruction(&wasora_instruction_dae, NULL)) == NULL) {
          return WASORA_PARSER_ERROR;
        }
        // y ponemos una finite-state machine en modo "leyendo daes"
        wasora_dae.reading_daes = 1;
      }

      if (matrix != NULL) {
        for (i = i_min; i < i_max; i++) {
          for (j = j_min; j < j_max; j++) {

            //wasora_define_instruction(wasora_instruction_dae, &wasora_dae.residual[wasora.i_dae]);

            sprintf(buffer, "%d", i+1);
            wasora_parse_expression(buffer, &wasora_dae.residual[wasora.i_dae].row);

            sprintf(buffer, "%d", j+1);
            wasora_parse_expression(buffer, &wasora_dae.residual[wasora.i_dae].col);

            wasora_dae.residual[wasora.i_dae].expr = malloc(sizeof(expr_t *));
            wasora_dae.residual[wasora.i_dae].expr[0] = malloc(sizeof(expr_t));
            if (wasora_parse_expression(equation, &wasora_dae.residual[wasora.i_dae].expr[0][0]) != 0) {
              return WASORA_PARSER_ERROR;
            }
            wasora.i_dae++;
          }
        }

      } else if (vector != NULL) {
        for (i = i_min; i < i_max; i++) {
          //wasora_define_instruction(wasora_instruction_dae, &wasora_dae.residual[wasora.i_dae]);

          sprintf(buffer, "%d", i+1);
          wasora_parse_expression(buffer, &wasora_dae.residual[wasora.i_dae].row);

          wasora_dae.residual[wasora.i_dae].expr = malloc(sizeof(expr_t *));
          wasora_dae.residual[wasora.i_dae].expr[0] = malloc(sizeof(expr_t));
          if (wasora_parse_expression(equation, &wasora_dae.residual[wasora.i_dae].expr[0][0]) != 0) {
            return WASORA_PARSER_ERROR;
          }
          wasora.i_dae++;
        }

      } else {

        //wasora_define_instruction(wasora_instruction_dae, &wasora_dae.residual[wasora.i_dae]);

        wasora_dae.residual[wasora.i_dae].expr = malloc(sizeof(expr_t *));
        wasora_dae.residual[wasora.i_dae].expr[0] = calloc(1, sizeof(expr_t));
        if (wasora_parse_expression(equation, &wasora_dae.residual[wasora.i_dae].expr[0][0]) != 0) {
          return WASORA_PARSER_ERROR;
        }

        wasora.i_dae++;

      }


      if (wasora.i_dae == wasora_dae.dimension) {
        // si es la ultima verificamos que no hubo instrucciones en el medio
        if (wasora.last_defined_instruction != wasora_dae.instruction) {
          wasora_push_error_message("cannot have instructinos within DAEs\n");
          return WASORA_PARSER_ERROR;
        }
        // y sacamos la finite-state machine del modo "leyendo daes"
        wasora_dae.reading_daes = 0;
      }

      free(equation);

      return WASORA_PARSER_OK;

    // ----- ASSIGNMENT -----------------------------------------------------------
    } else if ((token = wasora_get_next_token(NULL)) != NULL && strcmp(token, "=") == 0) {
///kw+=+usage { <var> | <vector>(<expr_i>) | <matrix>(<expr_i>,<expr_j>) [ [<expr_tmin>, <expr_tmax>] | @<expr_t> ]} = { <expr> }

      char *left_hand;
      char *right_hand;
      vector_t *vector = NULL;
      matrix_t *matrix = NULL;
      char *dummy_at;
      char *at_expression;
      int at_offset;
      char *fake_time_range;
      char *dummy_time_range;
      char *dummy_index_range;
      char *dummy_par;
      char *dummy_init;
      char *dummy_0;

      assignment_t *assignment;
      assignment = calloc(1, sizeof(assignment_t));
      LL_APPEND(wasora.assignments, assignment);

      left_hand = strdup(wasora.line);
      dummy = strchr(left_hand, '=');
      *dummy = '\0';

      // primero hacemos que el right hand sea dummy+1
      right_hand = strdup(dummy+1);
      wasora_strip_blanks(right_hand);
      if (strlen(right_hand) == 0) {
        wasora_push_error_message("empty right-hand side");
        return WASORA_PARSER_ERROR;
      }

      // y despues sacamos blancos y eso
      wasora_strip_blanks(left_hand);

      // este es el orden inverso en el que se tienen que dar los subindices
      // por ejemplo un construcciones validas serian
      // matrix_0(i,j)<1:3;2:4> = algo
      // matrix(i,j)<1:3;2:4>[0.99:1.01] = algo
      
      // como ultima cosa, la arroba para hacer steps
      if ((dummy_at = strchr(left_hand, '@')) != NULL) {
        at_offset = dummy_at - left_hand;
        at_expression = strdup(dummy_at+1);
        left_hand = realloc(left_hand, strlen(left_hand) + 2*strlen(at_expression) + 32);
        sprintf(left_hand+at_offset, "[(%s)-0.1*dt:%s+(0.1)*dt]", at_expression, at_expression);
        free(at_expression);
      }
      
      // como anteultima cosa, el rango en tiempo
      if ((dummy_time_range = strchr(left_hand, '[')) != NULL) {
        *dummy_time_range = '\0';
      }

      // el rango en indices
      if ((dummy_index_range = strchr(left_hand, '<')) != NULL) {
        *dummy_index_range = '\0';
      }

      // despues los subindices entre parentesis si es un vector o matriz
      if ((dummy_par = strchr(left_hand, '(')) != NULL) {
        *dummy_par = '\0';
      }

      // despues _init o _0
      if ((dummy_init = wasora_ends_in_init(left_hand)) != NULL) {
        *dummy_init = '\0';
        assignment->initial_static = 1;
      }

      if ((dummy_0 = wasora_ends_in_zero(left_hand)) != NULL) {
        *dummy_0 = '\0';
        assignment->initial_transient = 1;
      }

      if ((matrix = wasora_get_matrix_ptr(left_hand)) != NULL) {
        assignment->matrix = matrix;
      } else if ((vector = wasora_get_vector_ptr(left_hand)) != NULL) {
        assignment->vector = vector;
      } else {
        // buscamos una variable, y si no existe vemos si tenemos que declararla implicitamente
        if ((assignment->variable = wasora_get_variable_ptr(left_hand)) == NULL) {
          if (wasora.implicit_none) {
            wasora_push_error_message("undefined symbol '%s' and disabled implicit definition", left_hand);
            return WASORA_PARSER_ERROR;
          } else {
            if ((assignment->variable = wasora_define_variable(left_hand)) == NULL) {
              return WASORA_PARSER_ERROR;
            }
          }
        }
      }

      // ahora vamos de izquierda a derecha
      // reconstruimos el string
      if (dummy_0 != NULL) {
        *dummy_0 = '_';
      }
      if (dummy_init != NULL) {
        *dummy_init = '_';
      }

      // si hay parentesis son subindices
      if (dummy_par != NULL) {
        if (matrix == NULL && vector == NULL) {
          wasora_push_error_message("symbol '%s' cannot be subscripted", left_hand);
          return WASORA_PARSER_ERROR;
        }
        *dummy_par = '(';

        if (matrix != NULL) {
          if (wasora_parse_range(dummy_par, '(', ',', ')', &assignment->row, &assignment->col) != 0) {
            return WASORA_PARSER_ERROR;
          }
        } else if (vector != NULL) {
          if ((dummy = strrchr(dummy_par, ')')) == NULL) {
            wasora_push_error_message("unmatched parenthesis for '%s'", left_hand);
            return WASORA_PARSER_ERROR;
          }
          *dummy = '\0';
          if (wasora_parse_expression(dummy_par+1, &assignment->row) != 0) {
            return WASORA_PARSER_ERROR;
          }
          *dummy = ')';
        }
      }

      // si hay rango de indices
      if (dummy_index_range != NULL) {
        if (matrix == NULL && vector == NULL) {
          wasora_push_error_message("symbol '%s' cannot be subscripted", left_hand);
          return WASORA_PARSER_ERROR;
        }
        *dummy_index_range = '<';

        if (matrix != NULL) {
          if (wasora_parse_range(dummy_index_range, '<', ':', ';', &assignment->i_min, &assignment->i_max) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }
          if (wasora_parse_range(strchr(dummy_index_range, ';'), ';', ':', '>', &assignment->j_min, &assignment->j_max) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }
        } else if (vector != NULL) {
          if (wasora_parse_range(dummy_index_range, '<', ':', '>', &assignment->i_min, &assignment->i_max) != WASORA_PARSER_OK) {
            return WASORA_PARSER_ERROR;
          }
        }
      }

      // rango de tiempo
      if (dummy_time_range != NULL)  {
        *dummy_time_range = '[';
        if (wasora_parse_range(dummy_time_range, '[', ':', ']', &assignment->t_min, &assignment->t_max) != WASORA_PARSER_OK) {
          return WASORA_PARSER_ERROR;
        }
      }


      // evaluamos los extremos de los rangos
      // el extremo inferior es facil, si nos dieron algo lo evaluamos sino es uno
      if (assignment->i_min.n_tokens != 0) {
        wasora_var(wasora_special_var(i)) = wasora_evaluate_expression(&assignment->i_min);
      } else {
        wasora_var(wasora_special_var(i)) = 1;
      }

      if (assignment->j_min.n_tokens != 0) {
        wasora_var(wasora_special_var(j)) = wasora_evaluate_expression(&assignment->j_min);
      } else {
        wasora_var(wasora_special_var(j)) = 1;
      }

      if (assignment->row.n_tokens != 0) {
        assignment->row_low = wasora_evaluate_expression(&assignment->row);
      } else {
        assignment->row_low = 1;
      }

      if (assignment->col.n_tokens != 0) {
        assignment->col_low = wasora_evaluate_expression(&assignment->col);
      } else {
        assignment->col_low = 1;
      }


      // el extremo superior es si nos dieron algo lo evaluamos, sino depende del objeto
      if (assignment->i_min.n_tokens != 0) {
        wasora_var(wasora_special_var(i)) = wasora_evaluate_expression(&assignment->i_max);
      } else {
        if (assignment->matrix != NULL) {
          wasora_var(wasora_special_var(i)) = assignment->matrix->rows;
        } else if (assignment->vector != NULL) {
          wasora_var(wasora_special_var(i)) = assignment->vector->size;
        } else if (assignment->variable != NULL) {
          wasora_var(wasora_special_var(i)) = 1;
        }
      }

      if (assignment->j_min.n_tokens != 0) {
        wasora_var(wasora_special_var(j)) = wasora_evaluate_expression(&assignment->j_max);
      } else {
        if (assignment->matrix != NULL) {
          wasora_var(wasora_special_var(j)) = assignment->matrix->cols;
        } else {
          wasora_var(wasora_special_var(j)) = 1;
        }
      }

      if (assignment->row.n_tokens != 0) {
        assignment->row_high = wasora_evaluate_expression(&assignment->row);
      } else {
        // la unica manera de que no haya expresion para row es que sea una variable
        assignment->row_high = 1;
      }

      if (assignment->col.n_tokens != 0) {
        assignment->col_high = wasora_evaluate_expression(&assignment->col);
      } else {
        // la unica manera de que no haya expresion para row es que sea una variable o un vector
        assignment->col_high = 1;
      }

      if (assignment->row_high < assignment->row_low) {
        wasora_push_error_message("upper bound of row index is less than lower bound (%d and %d) in %s", assignment->row_high, assignment->row_low, left_hand);
        return WASORA_PARSER_ERROR;
      }
      if (assignment->col_high < assignment->col_low) {
        wasora_push_error_message("upper bound of column index is less than lower bound (%d and %d) in %s", assignment->col_high, assignment->col_low, left_hand);
        return WASORA_PARSER_ERROR;
      }


      assignment->expr = calloc(1+assignment->row_high-assignment->row_low, sizeof(expr_t *));
      for (i = assignment->row_low; i < assignment->row_high+1; i++) {
        assignment->expr[i-assignment->row_low] = calloc(1+assignment->col_high-assignment->col_low, sizeof(expr_t));
        for (j = assignment->col_low; j < assignment->col_high+1; j++) {
          if (wasora_parse_expression(right_hand, &assignment->expr[i-assignment->row_low][j-assignment->col_low]) != 0) {
            return WASORA_PARSER_ERROR;
          }
        }
      }

      free(left_hand);
      free(right_hand);

      if (wasora_define_instruction(wasora_instruction_assignment, assignment) == NULL) {
        return WASORA_PARSER_ERROR;
      }

      return WASORA_PARSER_OK;

    }


  }

  return WASORA_PARSER_UNHANDLED;

}


char *wasora_get_nth_token(char *string, int n) {

  char *backup;
  char *token;
  char *desired_token;
  int i;

  backup = strdup(string);

  if ((token = strtok(backup, UNQUOTED_DELIM)) == NULL) {
    free(backup);
    return NULL;
  }
  i = 1;

  while (i < n) {
    if ((token = strtok(NULL, UNQUOTED_DELIM)) == NULL) {
      free(backup);
      return NULL;
    }
    i++;
  }

  desired_token = strdup(token);
  free(backup);

  return desired_token;


}
